import pandas as pd
import pytest
from tradingenv.features import Feature
from tradingenv.events import Observer
import gym.spaces
from datetime import datetime
from tradingenv.events import IEvent
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from tradingenv.events import EventNBBO
from tradingenv.contracts import ETF
import numpy as np
# TODO: find an interface of features possibly based on sklearn
# TODO: feature should be optionally scaled using sklearn transformers, if
#  provided - no need to reinvent the wheel and use familiar API. There are
#  use cases where the user does not need to transform (e.g. backtesting) so
#  this should be a feature and not a necessary condition.
# TODO: AbstractFeature should implement reset no?


class TestFeaturesTransformers:
    def test_raises_if_transformer_is_passed_without_space(self):
        with pytest.raises(ValueError):
            Feature(transformer=StandardScaler())

    def test_reset_preserves_transformer_attributes(self):
        # https://stackoverflow.com/questions/55731933/initialise-standardscaler-from-scaling-parameters
        f = Feature(
            space=gym.spaces.Box(-100, +100, (1, 2)),
            transformer=StandardScaler(),
        )
        f.transformer.mean_ = np.array(([[1., 2.]]))
        f.reset()

        actual = f.transformer.mean_
        expected = np.array(([[1., 2.]]))
        np.testing.assert_equal(actual, expected)

    def test_transform_persist_across_reset(self):
        class FeaturePortfolioWeight(Feature):
            """Feature instancing the transformer in init, therefore not
            saved to _init_args/kwargs, make sure that transformer fit persists
            across resets."""

            def __init__(self, contracts):
                self.contracts = contracts
                super().__init__(
                    space=gym.spaces.Box(0., 1., (1, len(self.contracts))),
                    transformer=MinMaxScaler((-1, 1)),
                )

            def parse(self):
                return np.array([[0.5]])

            def _manual_fit_transformer(self):
                """Whatever the max or min allocation is per asset, either long or
                short, rescale the current portfolio weights in the [-1, +1] range."""
                x = np.concatenate([self.space.high, self.space.low])
                self.transformer.fit(x)

        feature = FeaturePortfolioWeight([ETF('SPY')])
        feature.reset()
        actual = feature()
        expected = np.array([[0.5]])
        np.testing.assert_almost_equal(actual, expected)
        feature.fit_transformer()

        actual = feature()
        expected = np.array([[0.]])
        np.testing.assert_almost_equal(actual, expected)

        feature.reset()

        actual = feature()
        expected = np.array([[0.]])
        np.testing.assert_almost_equal(actual, expected)

    @pytest.mark.skip(reason='TODO: Refactor')
    def test_can_override_parameters_of_StandardScaler(self):
        class MyFeature(Feature):
            def process_EventNBBO(self, event):
                self.event = event

            def parse(self):
                data = np.array([[self.event.bid_price, self.event.ask_price]])
                data *= np.random.normal()
                return data

        feature = MyFeature(
            space=gym.spaces.Box(-100, +100, (1, 2)),
            transformer=StandardScaler(),
        )
        feature.reset()
        for t in pd.date_range('2022-01-01', periods=10, freq='B'):
            event = EventNBBO(t, ETF('SPY'), 1., 1., 1., 1.)
            event.notify([feature])

        X = np.concatenate(list(feature.history.values()))
        scaler = StandardScaler()
        scaler.fit(X)
        mean, scale, var = scaler.mean_, scaler.scale_, scaler.var_
        expected = scaler.transform(X)

        scaler = StandardScaler()
        scaler.mean_, scaler.scale_, scaler.var_ = mean, scale, var
        actual = scaler.transform(X)
        np.testing.assert_equal(actual, expected)



"""https://nbconvert.readthedocs.io/en/latest/execute_api.html"""
import os
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert import HTMLExporter
import pytest
from pathlib import Path
here = Path(os.path.abspath(os.path.dirname(__file__)))


def ipynb_paths():
    paths = list()
    for root, dirs, files in os.walk(here):
        folder = os.path.basename(root)
        if not folder.startswith('.'):
            for file in files:
                if file.endswith(".ipynb") and not file.startswith('.'):
                    paths.append(os.path.join(root, file))
    return paths


class TestExamples:
    @pytest.mark.parametrize('path_to_ipynb', ipynb_paths())
    def test_notebook(self, path_to_ipynb):
        """goal is to load all .ipynb from /notebooks (v), run them (v)
        and then save them as html with toc.
        tutorial: https://nbconvert.readthedocs.io/en/latest/nbconvert_library.html

        Notes
        -----
        Rendering cufflinks charts in jupyter lab requires the extension
        described at the following link:
        https://medium.com/@hicraigchen/plotly-with-pandas-via-cufflinks-in-jupyter-lab-issues-50fcf1a89a1c
        """
        with open(path_to_ipynb) as f:
            nb = nbformat.read(f, as_version=4)

        # Configure the execution mode.
        ep = ExecutePreprocessor(timeout=-1, kernel_name='python3')

        # Execute/Run (preprocess):
        ep.preprocess(nb)

        # 2. Instantiate the exporter. We use the `basic` template for now;
        # we'll get into more details later about how to customize the
        # exporter further.
        html_exporter = HTMLExporter()
        html_as_string, resources = html_exporter.from_notebook_node(nb)

        # Save.
        dir_examples_html = os.path.join(here, 'html')
        os.makedirs(dir_examples_html, exist_ok=True)
        html_name = os.path.basename(path_to_ipynb).replace('ipynb', 'html')
        html_path = os.path.join(dir_examples_html, html_name)

        with open(html_path, 'w', encoding='utf-8') as f:
            f.write(html_as_string)



from tradingenv.broker.rebalancing import Rebalancing
from tradingenv.broker.fees import IBrokerFees, BrokerFees
from tradingenv.broker.trade import Trade
from tradingenv.broker.broker import (
    Context,
    Broker,
    EndOfEpisodeError,
    SECONDS_IN_YEAR,
)
from tradingenv.exchange import Exchange
from tradingenv.contracts import Cash, Rate, ETF, ES, VX
from tradingenv.events import EventNBBO
from datetime import datetime, timedelta
from typing import List
import numpy as np
import pytest


class TestBroker:
    def make_exchange(self, nbbo_events: List[EventNBBO] = None):
        if nbbo_events is None:
            nbbo_events = list()
        nbbo_events += [
            EventNBBO(datetime.now(), Cash(), 1, 1),
            EventNBBO(datetime.now(), Rate("FED funds rate"), 0.0, 0.0),
        ]
        exchange = Exchange()
        for nbbo in nbbo_events:
            exchange.process_EventNBBO(nbbo)
        return exchange

    def test_default_initialization(self):
        exchange = Exchange()
        broker = Broker(exchange)
        assert exchange is exchange
        assert broker.base_currency == Cash()
        assert broker.fees.interest_rate == Rate("FED funds rate")
        assert broker.holdings_quantity == {broker.base_currency: 100.0}
        assert isinstance(broker.fees, BrokerFees)
        assert broker.holdings_margins == dict()
        assert broker._last_accrual is None
        assert broker._initial_deposit == 100.0

    def test_custom_initialization(self):
        exchange = Exchange()
        broker = Broker(
            exchange,
            Cash("EUR"),
            42.0,
            BrokerFees(markup=0.1, interest_rate=Rate("Euribor")),
        )
        assert exchange is exchange
        assert broker.base_currency == Cash("EUR")
        assert broker.fees.interest_rate == Rate("Euribor")
        assert broker._last_accrual is None
        assert broker.fees.markup == 0.1
        assert broker.holdings_margins == dict()
        assert broker.holdings_quantity == {broker.base_currency: 42.0}
        assert broker._initial_deposit == 42.0

    def test_holdings_quantity_returns_copy(self):
        broker = Broker(self.make_exchange())
        first = broker.holdings_quantity
        second = broker.holdings_quantity
        assert first is not second

    def test_holdings_margins_returns_copy(self):
        broker = Broker(self.make_exchange())
        first = broker.holdings_margins
        second = broker.holdings_margins
        assert first is not second

    def test_holdings_values_raises_when_missing_prices(self):
        exchange = self.make_exchange()
        broker = Broker(exchange)
        broker._holdings_quantity[ETF("SPY")] = 1.0
        msg = "Missing liquidation transaction_price for"
        with pytest.raises(ValueError, match=msg):
            broker.holdings_values()

    def test_holdings_values_works_when_missing_non_needed_prices(self):
        exchange = self.make_exchange()
        broker = Broker(exchange)
        broker._holdings_quantity[ETF("SPY")] = 0.0
        actual = broker.holdings_values()
        expected = {broker.base_currency: broker._initial_deposit, ETF("SPY"): 0.0}
        assert actual, expected

    def test_holdings_values_when_price_spread_is_zero(self):
        exchange = self.make_exchange(
            nbbo_events=[
                EventNBBO(datetime.now(), ETF("SPY"), 2, 2),
                EventNBBO(datetime.now(), ETF("IEF"), 3, 3),
            ]
        )
        broker = Broker(exchange)
        broker._holdings_quantity[broker.base_currency] = 10.0
        broker._holdings_quantity[ETF("SPY")] = -42
        broker._holdings_quantity[ETF("IEF")] = 18
        actual = broker.holdings_values()
        expected = {
            broker.base_currency: 10 * 1,
            ETF("SPY"): -42 * 2,
            ETF("IEF"): 18 * 3,
        }
        assert actual == expected

    def test_holdings_values_when_price_spread_is_positive(self):
        exchange = self.make_exchange(
            nbbo_events=[
                EventNBBO(datetime.now(), ETF("SPY"), 1.99, 2.01),
                EventNBBO(datetime.now(), ETF("IEF"), 2.98, 3.02),
            ]
        )
        broker = Broker(exchange)
        broker._holdings_quantity[broker.base_currency] = 10.0
        broker._holdings_quantity[ETF("SPY")] = -42
        broker._holdings_quantity[ETF("IEF")] = 18
        actual = broker.holdings_values()
        expected = {
            broker.base_currency: 10 * 1,
            ETF("SPY"): -42 * 2.01,
            ETF("IEF"): 18 * 2.98,
        }
        assert actual == expected

    def test_holdings_values_is_multiplied_by_multiplier(self):
        exchange = self.make_exchange(
            nbbo_events=[EventNBBO(datetime.now(), ETF("SPY"), 2, 2),
                         EventNBBO(datetime.now(), ES(2019, 6), 3, 3)]
        )
        broker = Broker(exchange)
        broker._holdings_quantity[broker.base_currency] = 10.0
        broker._holdings_quantity[ETF("SPY")] = -42
        broker._holdings_quantity[ES(2019, 6)] = 18
        actual = broker.holdings_values()
        expected = {
            broker.base_currency: 10 * 1,
            ETF("SPY"): -42 * 2 * ETF("SPY").multiplier,
            ES(2019, 6): 18 * 3 * ES(2019, 6).multiplier,
        }
        assert actual == expected

    def test_net_liquidation_value(self):
        exchange = self.make_exchange(
            nbbo_events=[
                EventNBBO(datetime.now(), ETF("SPY"), 1.99, 2.01),
                EventNBBO(datetime.now(), ETF("IEF"), 2.98, 3.02),
            ]
        )
        broker = Broker(exchange)
        broker._holdings_quantity[broker.base_currency] = 10.0
        broker._holdings_quantity[ETF("SPY")] = -2
        broker._holdings_quantity[ETF("IEF")] = 18
        actual = broker.net_liquidation_value()
        expected = 10 * 1 - 2 * 2.01 + 18 * 2.98
        assert actual == expected

    def test_net_liquidation_value_raises_if_nlv_is_zero(self):
        exchange = self.make_exchange(
            nbbo_events=[
                EventNBBO(datetime.now(), ETF("SPY"), 1.99, 2.01),
                EventNBBO(datetime.now(), ETF("IEF"), 2.98, 3.02),
            ]
        )
        broker = Broker(exchange)
        broker._holdings_quantity[broker.base_currency] = 0.0
        broker._holdings_quantity[ETF("SPY")] = 0.0
        broker._holdings_quantity[ETF("IEF")] = 0.0
        with pytest.raises(EndOfEpisodeError):
            broker.net_liquidation_value()

    def test_net_liquidation_value_raises_if_nlv_is_negative(self):
        exchange = self.make_exchange(
            nbbo_events=[
                EventNBBO(datetime.now(), ETF("SPY"), 1.99, 2.01),
                EventNBBO(datetime.now(), ETF("IEF"), 2.98, 3.02),
            ]
        )
        broker = Broker(exchange)
        broker._holdings_quantity[broker.base_currency] = 1
        broker._holdings_quantity[ETF("SPY")] = -1
        broker._holdings_quantity[ETF("IEF")] = 0.0
        with pytest.raises(EndOfEpisodeError):
            broker.net_liquidation_value()

    def test_net_liquidation_value_when_marking_to_market_is_needed_1(self):
        exchange = self.make_exchange([EventNBBO(datetime.now(), ES(2019, 6), 99, 100)])
        broker = Broker(exchange, deposit=100.0)
        notional_value_of_one_contract = 100 * ES(2019, 6).multiplier
        quantity = broker._initial_deposit / notional_value_of_one_contract
        trade = Trade(datetime.now(), ES(2019, 6), quantity, 99, 100)
        broker.transact(trade)
        trade = Trade(datetime.now(), ES(2019, 6), -quantity, 99, 100)
        broker.transact(trade)
        # bid_ask spread deduced from cash
        assert broker.net_liquidation_value() == 99

    def test_net_liquidation_value_when_marking_to_market_is_needed_2(self):
        exchange = self.make_exchange([EventNBBO(datetime.now(), ES(2019, 6), 100, 100)])
        broker = Broker(exchange, deposit=100.0)
        notional_value_of_one_contract = 100 * ES(2019, 6).multiplier
        quantity = broker._initial_deposit / notional_value_of_one_contract
        trade = Trade(datetime.now(), ES(2019, 6), quantity, 100, 100)
        broker.transact(trade)
        exchange.process_EventNBBO(EventNBBO(datetime.now(), ES(2019, 6), 101, 101))
        # bid_ask spread deduced from cash
        assert broker.net_liquidation_value() == 101.0

    @pytest.mark.parametrize(
        "holdings_quantity, expected",
        [
            (
                {Cash(): 100, ETF("SPY"): 50, ETF("IEF"): 100 / 3},
                {Cash(): 1 / 3, ETF("SPY"): 1 / 3, ETF("IEF"): 1 / 3},
            ),
            (
                {Cash(): 200, ETF("SPY"): 100, ETF("IEF"): -200 / 3},
                {Cash(): 1, ETF("SPY"): 1, ETF("IEF"): -1},
            ),
            (
                {Cash(): 100, ETF("SPY"): 0, ETF("IEF"): 0},
                {Cash(): 1, ETF("SPY"): 0, ETF("IEF"): 0},
            ),
            (
                {Cash(): 1, ETF("SPY"): 0, ETF("IEF"): 0},
                {Cash(): 1, ETF("SPY"): 0, ETF("IEF"): 0},
            ),
            (
                {Cash(): -100, ETF("SPY"): 100, ETF("IEF"): 0},
                {Cash(): -1, ETF("SPY"): 2, ETF("IEF"): 0},
            ),
        ],
    )
    def test_holdings_weights(self, holdings_quantity, expected):
        exchange = self.make_exchange(
            nbbo_events=[
                EventNBBO(datetime.now(), ETF("SPY"), 2, 2),
                EventNBBO(datetime.now(), ETF("IEF"), 3, 3)]
        )
        broker = Broker(exchange)
        broker._holdings_quantity = holdings_quantity
        actual = broker.holdings_weights()
        assert actual == expected

    def test_holdings_weights_with_drifting_prices(self):
        exchange = self.make_exchange(
            nbbo_events=[EventNBBO(datetime.now(), ETF("SPY"), 2, 2),
                         EventNBBO(datetime.now(), ETF("IEF"), 3, 3)]
        )
        broker = Broker(exchange)
        assert broker.holdings_quantity == {broker.base_currency: 100.0}

        broker._holdings_quantity = {
            broker.base_currency: 30,
            ETF("SPY"): 20,
            ETF("IEF"): -10,
        }
        exchange.process_EventNBBO(EventNBBO(datetime.now(), ETF("SPY"), 1.99, 2.01))
        exchange.process_EventNBBO(EventNBBO(datetime.now(), ETF("IEF"), 2.98, 3.02))
        actual = broker.holdings_weights()
        holdings_values = {
            broker.base_currency: 30 * 1,
            ETF("SPY"): 20 * 1.99,
            ETF("IEF"): -10 * 3.02,
        }
        nlv = sum(holdings_values.values())
        expected = {
            broker.base_currency: 30 * 1 / nlv,
            ETF("SPY"): 20 * 1.99 / nlv,
            ETF("IEF"): -10 * 3.02 / nlv,
        }
        assert actual == expected

        exchange.process_EventNBBO(EventNBBO(datetime.now(), ETF("SPY"), 2.44, 2.48))
        exchange.process_EventNBBO(EventNBBO(datetime.now(), ETF("IEF"), 2.14, 2.31))
        actual = broker.holdings_weights()
        holdings_values = {
            broker.base_currency: 30 * 1,
            ETF("SPY"): 20 * 2.44,
            ETF("IEF"): -10 * 2.31,
        }
        nlv = sum(holdings_values.values())
        expected = {
            broker.base_currency: 30 * 1 / nlv,
            ETF("SPY"): 20 * 2.44 / nlv,
            ETF("IEF"): -10 * 2.31 / nlv,
        }
        assert actual == expected

    @pytest.mark.parametrize(
        "deposit, rate, markup, expected",
        [
            (1.0, 0, 0, 0),
            (1, 0.01, 0, 0.01),
            (1, 0.01, 0.01, 0),
            (-1, 0.01, 0, -0.01),
            (-1, 0.01, 0.01, -0.02),
            (1, 0.03, 0.01, 0.02),
            (-1, 0.03, 0.01, -0.04),
            (1, 0.05, 0.005, 0.045),
            (-1, 0.05, 0.005, -0.055),
            (1, -0.05, 0.005, 0),  # don't accrue negative interest rates
        ],
    )
    def test_accrued_interest(self, deposit, rate, markup, expected):
        exchange = Exchange()
        broker = Broker(exchange, deposit=deposit, fees=BrokerFees(markup))
        exchange.process_EventNBBO(EventNBBO(datetime.now(), broker.fees.interest_rate, rate, rate))
        broker._last_accrual = datetime(2019, 1, 1)
        now = broker._last_accrual + timedelta(seconds=SECONDS_IN_YEAR)

        # Assert value is correct...
        np.testing.assert_almost_equal(
            actual=broker.accrued_interest(now), desired=expected, decimal=7
        )
        # ...but not reflected in position holdings as accrue=False by default.
        assert broker.holdings_quantity[broker.base_currency] == deposit
        assert broker._last_accrual == datetime(2019, 1, 1)

        # Assert value is correct...
        np.testing.assert_almost_equal(
            actual=broker.accrued_interest(now, accrue=True),
            desired=expected,
            decimal=7,
        )
        # ...but not reflected in position holdings when accrue=True.
        assert broker.holdings_quantity[broker.base_currency] == deposit + expected
        assert broker._last_accrual == now

    def test_accrued_interest_sets_last_update_when_run_for_the_first_time(self):
        exchange = Exchange()
        broker = Broker(exchange)
        assert broker._last_accrual is None
        broker.accrued_interest(datetime(2019, 1, 1))
        assert broker._last_accrual == datetime(2019, 1, 1)

    def test_accrued_interest_raise_if_now_less_then_last_update_time(self):
        exchange = Exchange()
        broker = Broker(exchange)
        broker._last_accrual = datetime(2019, 1, 1)
        with pytest.raises(ValueError):
            broker.accrued_interest(datetime(2018, 6, 6))

    @pytest.mark.parametrize(
        "data",
        [
            {
                "trades": [
                    Trade(datetime.now(), ETF("SPY"), 1, 41, 43),
                    Trade(datetime.now(), ETF("IEF"), 2.5, 17, 19),
                ],
                "holdings_quantity_expected": {
                    Cash(): 100 - 1 * 43 - 2.5 * 19,
                    ETF("SPY"): 1,
                    ETF("IEF"): 2.5,
                },
                "holdings_margin_expected": {ETF("SPY"): 0.0, ETF("IEF"): 0.0},
            },
            {
                "trades": [
                    Trade(datetime.now(), ETF("SPY"), -1, 41, 43),
                    Trade(datetime.now(), ETF("IEF"), 2.5, 17, 19),
                ],
                "holdings_quantity_expected": {
                    Cash(): 100 + 1 * 41 - 2.5 * 19,
                    ETF("SPY"): -1,
                    ETF("IEF"): 2.5,
                },
                "holdings_margin_expected": {ETF("SPY"): 0.0, ETF("IEF"): 0.0},
            },
            {
                "trades": [
                    Trade(datetime.now(), ETF("SPY"), 1, 41, 43),
                    Trade(datetime.now(), ETF("IEF"), -2.5, 17, 19),
                ],
                "holdings_quantity_expected": {
                    Cash(): 100 - 1 * 43 + 2.5 * 17,
                    ETF("SPY"): 1,
                    ETF("IEF"): -2.5,
                },
                "holdings_margin_expected": {ETF("SPY"): 0.0, ETF("IEF"): 0.0},
            },
            {
                "trades": [
                    Trade(datetime.now(), ETF("SPY"), -1, 41, 43),
                    Trade(datetime.now(), ETF("IEF"), -2.5, 17, 19),
                ],
                "holdings_quantity_expected": {
                    Cash(): 100 + 1 * 41 + 2.5 * 17,
                    ETF("SPY"): -1,
                    ETF("IEF"): -2.5,
                },
                "holdings_margin_expected": {ETF("SPY"): 0.0, ETF("IEF"): 0.0},
            },
            {
                "trades": [Trade(datetime.now(), ES(2019, 6), 2.5, 41, 43)],
                "holdings_quantity_expected": {
                    Cash(): 100
                    - 2.5
                    * 43
                    * ES(2019, 6).multiplier
                    * ES(2019, 6).margin_requirement,
                    ES(2019, 6): 2.5,
                },
                "holdings_margin_expected": {
                    ES(2019, 6): 2.5
                    * 43
                    * ES(2019, 6).multiplier
                    * ES(2019, 6).margin_requirement
                },
            },
            {
                "trades": [Trade(datetime.now(), ES(2019, 6), -2.5, 41, 43)],
                "holdings_quantity_expected": {
                    Cash(): 100
                    - 2.5
                    * 41
                    * ES(2019, 6).multiplier
                    * ES(2019, 6).margin_requirement,
                    ES(2019, 6): -2.5,
                },
                # Note positive _margin when shorting.
                "holdings_margin_expected": {
                    ES(2019, 6): 2.5
                    * 41
                    * ES(2019, 6).multiplier
                    * ES(2019, 6).margin_requirement
                },
            },
        ],
    )
    def test_transact(self, data):
        exchange = self.make_exchange()
        broker = Broker(exchange, deposit=100.0)
        for trade in data["trades"]:
            broker.transact(trade)
        assert broker.holdings_quantity == data["holdings_quantity_expected"]
        assert broker.holdings_margins == data["holdings_margin_expected"]

    def test_transact_futures_lifecycle_without_bid_ask_spread(self):
        exchange = self.make_exchange([EventNBBO(datetime.now(), ES(2019, 6), 42, 42)])
        broker = Broker(exchange, deposit=100.0)
        trade = Trade(datetime.now(), ES(2019, 6), 2.5, 42, 42)
        broker.transact(trade)
        steps = 1000
        for _ in range(steps):
            trade = Trade(datetime.now(), ES(2019, 6), -2.5 / steps, 42, 42)
            broker.transact(trade)
        np.testing.assert_almost_equal(
            actual=broker.holdings_quantity[ES(2019, 6)], desired=0.0
        )
        np.testing.assert_almost_equal(
            actual=broker.holdings_margins[ES(2019, 6)], desired=0.0
        )

    def test_transact_futures_lifecycle_with_bid_ask_spread_single_lot(self):
        exchange = self.make_exchange([EventNBBO(datetime.now(), ES(2019, 6), 99, 100)])
        broker = Broker(exchange, deposit=100.0)
        notional_value_of_one_contract = 100 * ES(2019, 6).multiplier
        quantity = broker._initial_deposit / notional_value_of_one_contract
        trade = Trade(datetime.now(), ES(2019, 6), quantity, 99, 100)
        broker.transact(trade)
        trade = Trade(datetime.now(), ES(2019, 6), -quantity, 99, 100)
        broker.transact(trade)
        np.testing.assert_almost_equal(
            actual=broker.holdings_quantity[Cash()],
            desired=99,  # bid_ask spread deduced from cash
        )

    def test_transact_futures_lifecycle_with_bid_ask_spread_multiple_lots(self):
        exchange = self.make_exchange([EventNBBO(datetime.now(), ES(2019, 6), 99, 100)])
        broker = Broker(exchange, deposit=100.0)
        notional_value_of_one_contract = 100 * ES(2019, 6).multiplier
        quantity = broker._initial_deposit / notional_value_of_one_contract
        trade = Trade(datetime.now(), ES(2019, 6), quantity, 99, 100)
        broker.transact(trade)
        steps = 1000
        for _ in range(steps):
            trade = Trade(datetime.now(), ES(2019, 6), -quantity / steps, 99, 100)
            broker.transact(trade)
        np.testing.assert_almost_equal(
            actual=broker.holdings_quantity[Cash()],
            desired=99,  # bid_ask spread deduced from cash
        )

    def test_transact_deduces_cost_of_commissions(self):
        class Fees(IBrokerFees):
            def commissions(self, trade) -> float:
                return 18

        fees = Fees()
        exchange = self.make_exchange([EventNBBO(datetime.now(), ETF("SPY"), 42, 42)])
        broker = Broker(exchange, deposit=100.0)
        trade = Trade(datetime.now(), ETF("SPY"), 2.5, 42, 42, fees)
        broker.transact(trade)
        trade = Trade(datetime.now(), ETF("SPY"), -2.5, 42, 42, fees)
        broker.transact(trade)
        np.testing.assert_almost_equal(
            actual=broker.holdings_quantity[ETF("SPY")], desired=0.0
        )
        np.testing.assert_almost_equal(
            actual=broker.holdings_margins[ETF("SPY")], desired=0.0
        )
        np.testing.assert_almost_equal(
            actual=broker.holdings_quantity[Cash()],
            desired=100 - 18 * 2,  # bid_ask spread deduced from cash
        )

    @pytest.mark.parametrize("arg", [ES(2019, 6), None])
    @pytest.mark.parametrize(
        "bid_prices, ask_prices",
        [
            # No bid-ask spread.
            [[1872.5, 1847.5, 1900.0, 1872.5], [1872.5, 1847.5, 1900.0, 1872.5]],
            # Bid-ask spread.
            [[1872.0, 1847.0, 1899.5, 1872.0], [1873.0, 1848.0, 1900.5, 1873.0]],
        ],
    )
    def test_marking_to_market(self, arg, bid_prices, ask_prices):
        es = ES(2019, 6)
        exchange = self.make_exchange()
        exchange.process_EventNBBO(EventNBBO(datetime.now(), es, bid_prices[0], ask_prices[0]))
        deposit = 100
        broker = Broker(exchange, deposit=deposit)

        # Buy ES.
        lob = exchange[es]
        trade = Trade(datetime.now(), es, 0.002, lob.bid_price, lob.ask_price)
        broker.transact(trade)

        # Check status.
        initial_margin = (
            lob.liq_price(trade.quantity)
            * es.margin_requirement
            * trade.quantity
            * es.multiplier
        )
        roundtrip_spread = trade.quantity * es.multiplier * lob.spread
        cash = broker.holdings_quantity[broker.base_currency]
        assert broker.holdings_margins[es] == initial_margin
        assert broker.holdings_quantity[es] == trade.quantity
        np.testing.assert_almost_equal(
            actual=broker.holdings_quantity[broker.base_currency],
            desired=deposit - initial_margin - roundtrip_spread,
        )
        np.testing.assert_almost_equal(
            actual=sum(broker.holdings_margins.values()) + cash,
            desired=deposit - roundtrip_spread,
        )

        # ES price moves
        exchange.process_EventNBBO(EventNBBO(datetime.now(), es, bid_prices[1], ask_prices[1]))

        # Check status. Nothing should have changed because mark-to-market
        # has not been called (yet).
        assert broker.holdings_margins[es] == initial_margin
        cash = broker.holdings_quantity[broker.base_currency]
        assert broker.holdings_quantity[es] == trade.quantity
        np.testing.assert_almost_equal(
            actual=broker.holdings_quantity[broker.base_currency],
            desired=deposit - initial_margin - roundtrip_spread,
        )
        np.testing.assert_almost_equal(
            actual=sum(broker.holdings_margins.values()) + cash,
            desired=deposit - roundtrip_spread,
        )

        # Check status. Something should have changed now that mark-to-market
        # has been called. Prices have moved down, so loss (<).
        broker.marking_to_market(arg)
        new_margin = (
            lob.liq_price(trade.quantity)
            * es.margin_requirement
            * trade.quantity
            * es.multiplier
        )
        capital_gain1 = es.multiplier * (bid_prices[1] - bid_prices[0]) * trade.quantity
        assert new_margin < initial_margin
        cash = broker.holdings_quantity[broker.base_currency]
        np.testing.assert_almost_equal(
            actual=broker.holdings_margins[es], desired=new_margin
        )
        assert broker.holdings_quantity[es] == trade.quantity
        np.testing.assert_almost_equal(
            actual=sum(broker.holdings_margins.values()) + cash,
            desired=deposit + capital_gain1 - roundtrip_spread,
        )
        np.testing.assert_almost_equal(
            actual=cash, desired=deposit - new_margin + capital_gain1 - roundtrip_spread
        )

        # ES price moves
        exchange.process_EventNBBO(EventNBBO(datetime.now(), es, bid_prices[2], ask_prices[2]))

        # Check status. Nothing should have changed because mark-to-market
        # has not been called (yet).
        assert new_margin < initial_margin
        np.testing.assert_almost_equal(
            actual=broker.holdings_margins[es], desired=new_margin
        )
        assert broker.holdings_quantity[es] == trade.quantity
        np.testing.assert_almost_equal(
            actual=sum(broker.holdings_margins.values()) + cash,
            desired=deposit + capital_gain1 - roundtrip_spread,
        )
        np.testing.assert_almost_equal(
            actual=broker.holdings_quantity[broker.base_currency],
            desired=deposit - new_margin + capital_gain1 - roundtrip_spread,
        )

        # Check status. Something should have changed now that mark-to-market
        # has been called. Prices have moved down, so loss (<).
        broker.marking_to_market(arg)
        previous_margin = new_margin
        new_margin = (
            lob.liq_price(trade.quantity)
            * es.margin_requirement
            * trade.quantity
            * es.multiplier
        )
        capital_gain2 = es.multiplier * (bid_prices[2] - bid_prices[1]) * trade.quantity
        assert new_margin > previous_margin
        cash = broker.holdings_quantity[broker.base_currency]
        np.testing.assert_almost_equal(
            actual=broker.holdings_margins[es], desired=new_margin
        )
        assert broker.holdings_quantity[es] == trade.quantity
        np.testing.assert_almost_equal(
            actual=sum(broker.holdings_margins.values()) + cash,
            desired=deposit + capital_gain1 + capital_gain2 - roundtrip_spread,
        )
        np.testing.assert_almost_equal(
            actual=cash,
            desired=deposit
            - new_margin
            + capital_gain1
            + capital_gain2
            - roundtrip_spread,
        )

        # ES price moves and close position. Price is start price, so PnL
        # should be zero.
        exchange.process_EventNBBO(EventNBBO(datetime.now(), es, bid_prices[3], ask_prices[3]))
        lob = exchange[es]
        trade = Trade(datetime.now(), es, -trade.quantity, lob.bid_price, lob.ask_price)
        broker.transact(trade)

        # Check status. Something should have changed now that mark-to-market
        # has been called. Prices have moved down, so loss (<).
        broker.marking_to_market(arg)
        capital_gain3 = es.multiplier * (bid_prices[3] - bid_prices[2]) * trade.quantity
        cash = broker.holdings_quantity[broker.base_currency]
        assert broker.holdings_margins[es] == 0
        assert broker.holdings_quantity[es] == 0
        np.testing.assert_almost_equal(actual=cash, desired=deposit - roundtrip_spread)

    def test_marking_to_market_does_nothing_when_prices_do_not_change(self):
        es = ES(2019, 6)
        exchange = self.make_exchange()
        exchange.process_EventNBBO(EventNBBO(datetime.now(), es, 1872.5, 1872.5))
        deposit = 100
        broker = Broker(exchange, deposit=deposit)

        # Buy ES.
        lob = exchange[es]
        trade = Trade(datetime.now(), es, 0.002, lob.bid_price, lob.ask_price)
        broker.transact(trade)

        # Check status.
        initial_margin = (
            lob.mid_price * es.margin_requirement * trade.quantity * es.multiplier
        )
        cash = broker.holdings_quantity[broker.base_currency]
        assert broker.holdings_margins[es] == initial_margin
        assert broker.holdings_quantity[es] == trade.quantity
        assert (
            broker.holdings_quantity[broker.base_currency] == deposit - initial_margin
        )
        assert sum(broker.holdings_margins.values()) + cash == deposit

        # ES price moves
        exchange.process_EventNBBO(EventNBBO(datetime.now(), es, 1872.5, 1872.5))

        # Check status. Nothing should have changed because mark-to-market
        # has not been called.
        assert broker.holdings_margins[es] == initial_margin
        cash = broker.holdings_quantity[broker.base_currency]
        assert broker.holdings_quantity[es] == trade.quantity
        assert cash == deposit - initial_margin
        assert sum(broker.holdings_margins.values()) + cash == deposit

        # Check status. Nothing should have changed because prices haven't
        # changed.
        broker.marking_to_market()
        new_margin = (
            lob.mid_price * es.margin_requirement * trade.quantity * es.multiplier
        )
        capital_gain1 = es.multiplier * (1872.5 - 1872.5) * trade.quantity
        cash = broker.holdings_quantity[broker.base_currency]
        assert broker.holdings_margins[es] == new_margin
        assert broker.holdings_quantity[es] == trade.quantity
        assert sum(broker.holdings_margins.values()) + cash == deposit + capital_gain1
        assert cash == deposit - new_margin + capital_gain1

    def test_marking_to_market_does_not_raise_with_spot_products(self):
        spy = ETF("SPY")
        exchange = self.make_exchange()
        exchange.process_EventNBBO(EventNBBO(datetime.now(), spy, 1872.5, 1872.5))
        broker = Broker(exchange, deposit=100)
        lob = exchange[spy]
        trade = Trade(datetime.now(), spy, 0.002, lob.bid_price, lob.ask_price)
        broker.transact(trade)
        broker.marking_to_market(spy)  # does not raise

    @pytest.mark.parametrize('quantity', [-0.002, 0.002])
    def test_marking_to_market_margin_is_agnostic_to_sign_of_position(self, quantity):
        es = ES(2019, 3)
        exchange = self.make_exchange()
        exchange.process_EventNBBO(EventNBBO(datetime.now(), es, 1872.5, 1872.5))
        broker = Broker(exchange, deposit=100)
        lob = exchange[es]
        trade = Trade(datetime.now(), es, quantity, lob.bid_price, lob.ask_price)
        broker.transact(trade)
        broker.marking_to_market(es)
        assert broker.holdings_margins[es] == 1872.5 * abs(quantity) * es.multiplier * es.margin_requirement

    def test_spread_is_calculated(self):
        # Initial portfolio is $100 in cash. Target portfolio is 60:40 in
        # equities and bonds, both with a 0.02% of spread everywhere. So LOB is
        # 99.99<->100.01. So roundtrip cost in PnL is $100 to buy at 100.01
        # means that I'll own 100/100.01 shares to be sold at 99.99
        exchange = self.make_exchange()
        spread = 0.0002

        mid_price = 2
        halftrip_spread = mid_price * spread / 2
        exchange.process_EventNBBO(
            EventNBBO(
                datetime.now(),
                contract=ETF("SPY"),
                bid_price=mid_price - halftrip_spread,
                ask_price=mid_price + halftrip_spread,
            )
        )

        mid_price = 3
        halftrip_spread = mid_price * spread / 2
        exchange.process_EventNBBO(
            EventNBBO(
                datetime.now(),
                contract=ETF("IEF"),
                bid_price=mid_price - halftrip_spread,
                ask_price=mid_price + halftrip_spread,
            )
        )
        broker = Broker(exchange)

        request = Rebalancing(
            time=datetime(2019, 1, 6),
            contracts=[Cash(), ETF("SPY"), ETF("IEF")],
            allocation=[0, 0.75, 0.25],
        )
        broker.rebalance(request)
        np.testing.assert_approx_equal(
            actual=request.context_post.nlv,
            desired=broker._initial_deposit / 100.01 * 99.99
        )
        assert broker.holdings_quantity == {
            Cash(): 0,
            ETF("SPY"): broker._initial_deposit * 0.75 / exchange[ETF("SPY")].ask_price,
            ETF("IEF"): broker._initial_deposit * 0.25 / exchange[ETF("IEF")].ask_price,
        }

    def test_context(self):
        exchange = self.make_exchange(
            nbbo_events=[
                EventNBBO(datetime.now(), ETF("SPY"), 1.99, 2.01),
                EventNBBO(datetime.now(), ETF("IEF"), 2.98, 3.02),
            ]
        )
        broker = Broker(exchange)
        broker._holdings_quantity[broker.base_currency] = 10.0
        broker._holdings_quantity[ETF("SPY")] = -2
        broker._holdings_quantity[ETF("IEF")] = 18
        context = broker.context()
        assert context.nlv == broker.net_liquidation_value()
        assert context.nr_contracts == broker.holdings_quantity
        assert context.weights == broker.holdings_weights()
        assert context.margins == broker.holdings_margins
        assert context.values == broker.holdings_values()


class TestBrokerRebalance:
    def make_broker(self, deposit: float=100):
        nbbos = [
            EventNBBO(datetime.now(), Cash(), 1, 1),
            EventNBBO(datetime.now(), Rate("FED funds rate"), 0.0182, 0.0182),
            EventNBBO(datetime.now(), ETF('SPY'), 293.24, 293.25),
            EventNBBO(datetime.now(), ETF('IEF'), 112.67, 112.68),
            EventNBBO(datetime.now(), VX(2020, 3), 19.20, 19.25),
            EventNBBO(datetime.now(), ES(2020, 3), 2941.75, 2942),
            EventNBBO(datetime.now(), ES(2020, 6), 2940, 2940.25),
            EventNBBO(datetime.now(), ES(2020, 9), 2938.75, 2939),
            EventNBBO(datetime.now(), ES(2020, 12), 2935.25, 2935),
            EventNBBO(datetime.now(), ETF("Equities"), 2, 2),
            EventNBBO(datetime.now(), ETF("Bonds"), 3, 3),

        ]
        exchange = Exchange()
        for nbbo in nbbos:
            exchange.process_EventNBBO(nbbo)
        return Broker(exchange=exchange, base_currency=Cash(), deposit=deposit)

    def test_rebalance_interest_rate_on_idle_cash_is_accrued(self):
        tic = datetime(2019, 1, 1)
        toc = tic + timedelta(days=252)
        broker = self.make_broker(deposit=100)
        broker._last_accrual = tic

        assert broker.net_liquidation_value() == 100
        expected = broker.accrued_interest(toc, accrue=False)
        rebalancing = Rebalancing(time=toc)
        broker.rebalance(rebalancing)
        # we are not testing how much, but check that something is accrued.
        assert expected > 0
        assert rebalancing.profit_on_idle_cash == expected
        assert broker.net_liquidation_value() == 100 + expected

    def test_rebalance_rebalancing_object_is_added_to_track_record(self):
        broker = self.make_broker()
        rebalancing = Rebalancing()
        broker.rebalance(rebalancing)
        assert broker.track_record._rebalancing[rebalancing.time] is rebalancing

    def test_rebalance_trades_are_saved_to_attribute_in_rebalancing(self):
        broker = self.make_broker()
        rebalancing = Rebalancing([ETF('SPY')], [0.8])
        expected = rebalancing.make_trades(broker)
        assert not isinstance(rebalancing.trades, list)
        broker.rebalance(rebalancing)
        assert rebalancing.trades == expected

    def test_rebalance_context_pre(self):
        broker = self.make_broker()
        rebalancing = Rebalancing([ETF('SPY')], [0.8])
        expected = broker.context()
        assert not isinstance(rebalancing.context_pre, Context)
        broker.rebalance(rebalancing)
        assert rebalancing.context_pre == expected

    def test_rebalance_context_post(self):
        broker = self.make_broker()
        rebalancing = Rebalancing([ETF('SPY')], [0.8])
        assert not isinstance(rebalancing.context_post, Context)
        broker.rebalance(rebalancing)
        expected = broker.context()
        assert rebalancing.context_post == expected

    def test_rebalance_simplest_case_from_all_cash_to_all_cash(self):
        broker = self.make_broker(deposit=100)
        time = datetime.now()
        request = Rebalancing(
            contracts=[Cash(), ETF("SPY"), ETF("IEF")],
            allocation=[1, 0, 0],
            time=time,
        )
        broker.rebalance(request)

        assert broker.holdings_quantity == {Cash(): broker._initial_deposit}
        assert broker.net_liquidation_value() == broker._initial_deposit
        assert broker.holdings_weights() == {Cash(): 1}

    def test_rebalance_from_all_cash_to_balanced_portfolio(self):
        broker = self.make_broker(deposit=100)
        time = datetime.now()
        rebalancing = Rebalancing(
            contracts=[Cash(), ETF("Equities"), ETF("Bonds")],
            allocation=[0, 0.75, 0.25],
            time=time,
        )
        broker.rebalance(rebalancing)
        assert broker.holdings_quantity == {
            Cash(): 0,
            ETF("Equities"): broker._initial_deposit * 0.75 / 2,
            ETF("Bonds"): broker._initial_deposit * 0.25 / 3,
        }
        assert broker.net_liquidation_value() == broker._initial_deposit
        assert broker.holdings_weights() == {Cash(): 0, ETF('Equities'): 0.75, ETF('Bonds'): 0.25}



"""
We write examples in the test suite so that we don't have to use doctests.
Doctests are handy for simple and limited examples, but have their own issues
such as: (1) cannot correctly run or detect doctests separated by text or (2)
annoying clutter (>>> or ...) that hinders readability.

References
----------
https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-literalinclude
https://stackoverflow.com/questions/77172274
"""


"""TrackRecord stores historical _rebalancing execution performed by Broker."""
from tradingenv.broker.rebalancing import Rebalancing
from typing import Dict, Union, Sequence
from datetime import datetime
import pandas as pd
import numpy as np
import pickle
import plotly.graph_objs as go
import plotly.figure_factory
from pathlib import Path
import os


class TrackRecord:
    """Stores historical _rebalancing requests and _rebalancing."""

    def __init__(self):
        self._time = list()
        self._rebalancing: Dict[datetime, Rebalancing] = dict()
        self._nr_steps_to_burn = 0
        self._trading_started = False

    def __getitem__(self, item: Union[int, datetime]) -> Rebalancing:
        """If item is an int, the item-th RebalancingResponse will be returned.
        If item is datetime, the report associated with the specified time
        will be returned."""
        if isinstance(item, datetime):
            return self._rebalancing[item]
        else:
            return self._rebalancing[self._time[item]]

    def __len__(self):
        """Returns the count of historical checkpoints."""
        return len(self._rebalancing)

    def __repr__(self):
        return "{cls}({start}:{end}; items={nr_items})" "".format(
            cls=self.__class__.__name__,
            start=self._time[0],
            end=self._time[-1],
            nr_items=len(self._time),
        )

    def _checkpoint(self, rebalancing: Rebalancing):
        """Store an instance of RebalancingResponse. This class will take
        care of parsing the content using other methods such as
        TrackRecord.to_frame."""
        time = rebalancing.time
        if isinstance(time, pd.Timestamp):
            time = time.to_pydatetime()
        if time in self._time:
            raise ValueError(
                "All RebalancingResponse must have different timestamps. "
                "Duplicated timestamp found: {}".format(time)
            )
        self._time.append(time)
        self._rebalancing[time] = rebalancing
        self._trading_started = self._trading_started or len(rebalancing.trades) != 0
        if not self._trading_started:
            self._nr_steps_to_burn += 1

    def save(self, directory: str, name: str=None):
        # Use cloudpickle instead?
        os.makedirs(directory, exist_ok=True)
        file_name = name or self.name + '.pickle'
        path = Path(directory).joinpath(file_name)
        with open(str(path), "wb") as f:
            pickle.dump(self, f, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls, path: str) -> 'TrackRecord':
        # Use cloudpickle instead?
        with open(path, "rb") as f:
            track_record = pickle.load(f)
        return track_record

    def net_liquidation_value(
            self,
            before_rebalancing=True,
            burn: bool=False,
            name: str='Net liquidation value',
            index_name: str='Date',
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        before_rebalancing
        burn : bool
            Default is False (least surprise principle). If True, initial
            checkpoints with no trades (i.e. when the portfolio is cash only)
            will be discarded when calling TrackRecord.to_frame, unless
            a burn argument is explicitly passed to TrackRecord.to_frame.
        name
        index_name

        Returns
        -------

        """
        data = dict()
        for time, rebalancing in self._rebalancing.items():
            if before_rebalancing:
                context = rebalancing.context_pre
            else:
                context = rebalancing.context_post
            data[time] = context.nlv
        series = pd.Series(data=data, name=name, dtype=np.float64)
        series.index.name = index_name
        if burn:
            series = series.iloc[self._nr_steps_to_burn:]
        return series.to_frame()

    def weights_target(
            self,
            burn: bool=False,
            name: str='Portfolio weights target',
            index_name: str='Date',
            aggregate_future_chain: bool=False,
    ) -> pd.DataFrame:
        data = dict()
        for time, rebalancing in self._rebalancing.items():
            data[time] = rebalancing.allocation
        df = pd.DataFrame(data=data, dtype=np.float32).T
        df.index.name = index_name
        df.columns.name = name
        if burn:
            df = df.iloc[self._nr_steps_to_burn:]
        if aggregate_future_chain:
            groups = [c.symbol_short for c in df.columns]
            if len(groups) != len(df.columns):
                df = df.groupby(groups, axis='columns').sum()
        return df

    def weights_actual(
            self,
            before_rebalancing=True,
            burn: bool=False,
            name: str='Portfolio weights actual',
            index_name: str='Date',
            aggregate_future_chain: bool = False,
    ) -> pd.DataFrame:
        data = dict()
        for time, rebalancing in self._rebalancing.items():
            if before_rebalancing:
                context = rebalancing.context_pre
            else:
                context = rebalancing.context_post
            data[time] = context.weights
        df = pd.DataFrame(data=data, dtype=np.float32).T
        df.index.name = index_name
        df.columns.name = name
        if burn:
            df = df.iloc[self._nr_steps_to_burn:]
        if aggregate_future_chain:
            groups = [c.symbol_short for c in df.columns]
            if len(groups) != len(df.columns):
                df = df.groupby(groups, axis='columns').sum()
        return df

    def transaction_costs(
            self,
            burn: bool=False,
            index_name: str='Date',
            cumulative: bool=True,
    ) -> pd.DataFrame:
        profit_on_idle_cash = dict()
        cost_of_spread = dict()
        cost_of_commissions = dict()
        for time, rebalancing in self._rebalancing.items():
            profit_on_idle_cash[time] = rebalancing.profit_on_idle_cash
            cost_of_spread[time] = sum(trade.cost_of_spread for trade in rebalancing.trades)
            cost_of_commissions[time] = sum(trade.cost_of_commissions for trade in rebalancing.trades)
        df = pd.concat(
            objs=[
                pd.Series(profit_on_idle_cash).to_frame('Profit on idle Cash'),
                pd.Series(cost_of_spread).to_frame('Spread'),
                pd.Series(cost_of_commissions).to_frame('Broker fees'),
            ],
            axis='columns',
        )
        df.index.name = index_name
        if burn:
            df = df.iloc[self._nr_steps_to_burn:]
        if cumulative:
            df = df.cumsum().join(self.net_liquidation_value())
        return df

    def tearsheet(self):
        try:
            risk_free = self.risk_free
        except AttributeError:
            risk_free = 0
        try:
            benchmark = self.benchmark
        except AttributeError:
            benchmark = None
        nlv = self.net_liquidation_value()
        tearsheet = nlv.tearsheet(
            risk_free=risk_free,
            benchmark=benchmark,
            weights=self.weights_target(),
        )
        return tearsheet

    def to_excel(self, path):
        """Export track record to excel."""
        nlv = self.net_liquidation_value()
        nlv.index = nlv.index.date
        nlv.index.name = 'Date'
        weights = self.weights_target(aggregate_future_chain=True)
        weights.index = weights.index.date
        weights.index.name = 'Date'
        costs = self.transaction_costs()
        costs.index = costs.index.date
        costs.index.name = 'Date'
        with pd.ExcelWriter(path) as excel:
            nlv.to_excel(excel, 'Performance')
            weights.to_excel(excel, 'Weights')
            costs.to_excel(excel, 'Costs')

    def cost_of_spread(self):
        raise NotImplementedError()

    def cost_of_commissions(self):
        raise NotImplementedError()
    
    def _get_df_performance(self):
        nlv = self.net_liquidation_value()
        risk_free = self.risk_free.loc[nlv.index]
        benchmark = self.benchmark.loc[nlv.index]
        relative = (nlv.squeeze() / benchmark.squeeze()).to_frame(
            'Outperformance')
        df = pd.concat([nlv, risk_free, benchmark, relative], axis='columns')
        base_date = df.first_valid_index()
        df /= df.loc[base_date]
        df *= 100
        return df

    def fig_net_liquidation_value(self) -> go.Figure:
        df = self._get_df_performance()
        layout = go.Layout(
            title=go.layout.Title(
                text='Performance of {}'.format(self.name),
            ),
            yaxis=go.layout.YAxis(
                type='log',
                title=go.layout.yaxis.Title(text="Level")
            ),
        )
        data = list()
        for col in df.columns:
            scatter = go.Scatter(
                x=df.index,
                y=df[col],
                mode='lines',
                name=str(col),
                line=go.scatter.Line(width=2),
            )
            data.append(scatter)
        return go.Figure(data=data, layout=layout)

    def fig_drawdown(self) -> go.Figure:
        df = self._get_df_performance()
        df = df.drawdown()
        df *= 100

        # Layout.
        layout = go.Layout(
            title=go.layout.Title(
                text='Drawdown',
            ),
            yaxis=go.layout.YAxis(
                title=go.layout.yaxis.Title(text="%")
            ),
        )

        # Data.
        data = list()
        for col in df.columns:
            scatter = go.Scatter(
                x=df.index,
                y=df[col],
                mode='lines',
                name=str(col),
                line=go.scatter.Line(width=2),
            )
            data.append(scatter)
        return go.Figure(data=data, layout=layout)

    def fig_annual_returns(self) -> go.Figure:
        df = self._get_df_performance()

        annual_rets = df.resample("Y").last() / df.resample("Y").first() - 1
        annual_rets.index = annual_rets.index.year
        annual_rets *= 100

        # Layout.
        layout = go.Layout(
            title=go.layout.Title(
                text='Annual returns of {}'.format(self.name),
            ),
            yaxis=go.layout.YAxis(
                title=go.layout.yaxis.Title(text="%")
            ),
        )

        # Data.
        data = list()
        for col in annual_rets.columns:
            scatter = go.Bar(
                x=annual_rets.index,
                y=annual_rets[col],
                name=str(col),
            )
            data.append(scatter)
        return go.Figure(data=data, layout=layout)

    def fig_transaction_costs(self) -> go.Figure:
        df = self.transaction_costs(cumulative=True)
        layout = go.Layout(
            title=go.layout.Title(
                text='Transaction costs',
            ),
            yaxis=go.layout.YAxis(
                title=go.layout.yaxis.Title(text="$")
            ),
        )
        data = list()
        for col in df.columns:
            scatter = go.Scatter(
                x=df.index,
                y=df[col],
                mode='lines',
                name=str(col),
                line=go.scatter.Line(width=2),
            )
            data.append(scatter)
        return go.Figure(data=data, layout=layout)

    def fig_returns_distribution(self):
        nlv = pd.concat(
            objs=[
                self.net_liquidation_value(),
                self.benchmark,
            ],
            axis='columns',
            sort=False,
        )
        nlv.columns = [str(col) for col in nlv.columns]
        nlv.dropna(inplace=True)
        ret = nlv.log_returns()
        ret *= 100
        layout = go.Layout(
            title=go.layout.Title(text="Kernel Density Estimation"),
            yaxis=go.layout.YAxis(
                title=go.layout.yaxis.Title(text="Density")
            ),
            xaxis=go.layout.XAxis(
                title=go.layout.xaxis.Title(text="Return (%)")
            ),
        )
        data = plotly.figure_factory.create_distplot(
            hist_data=[ret[col] for col in ret.columns],
            group_labels=ret.columns,
            show_hist=False,
            show_curve=True,
            show_rug=False,
            curve_type="kde",
            histnorm="probability density",
        ).data
        return go.Figure(data=data, layout=layout)

    def fig_capital_asset_pricing_model(self) -> go.Figure:
        nlv = self.net_liquidation_value()
        capm = nlv.capm(
            benchmark=self.benchmark,
            risk_free=self.risk_free,
        )
        alpha = capm.params['alpha']
        beta = capm.params['beta']
        x = capm.model.data.exog[:, 0] * 100
        y = capm.model.data.endog * 100
        trendline = capm.fittedvalues * 100
        layout = go.Layout(
            title=go.layout.Title(
                text="Capital Asset Pricing Model<br>α={alpha:.2%} (annualized), β={beta:.2f}".format(
                    alpha=alpha, beta=beta),
            ),
            yaxis=go.layout.YAxis(
                title=go.layout.yaxis.Title(
                    text="Return (%) of the strategy over cash")
            ),
            xaxis=go.layout.XAxis(
                title=go.layout.xaxis.Title(
                    text="Return (%) of {} over cash".format(
                        self.benchmark.squeeze().name))
            ),
        )
        data = [
            go.Scatter(
                x=x,
                y=y,
                text=[t.date() for t in nlv.index],
                mode="markers",
                name="Returns",
                marker=go.scatter.Marker(size=5),
            ),
            go.Scatter(
                x=x,
                y=trendline,
                mode="lines",
                name="Trendline",
                showlegend=True,
            ),
        ]
        return go.Figure(data=data, layout=layout)

    def _historical_portfolio_weights(
            self, df: pd.DataFrame, title: str
    ) -> go.Figure:
        df *= 100
        layout = go.Layout(
            title=go.layout.Title(
                text=title,
            ),
            yaxis=go.layout.YAxis(
                title=go.layout.yaxis.Title(text="%")
            ),
        )
        data = list()
        for col, series in df.items():
            series = series.replace(0, np.nan).dropna()
            scatter = go.Scatter(
                x=series.index,
                y=series,
                mode='lines',
                name=str(col),
                line=go.scatter.Line(width=2),
            )
            data.append(scatter)
        return go.Figure(data=data, layout=layout)

    def fig_historical_portfolio_weights_actual(
            self, group_futures: bool=True, post_rebalancing: bool=True
    ) -> go.Figure:
        df = self.weights_actual(
            aggregate_future_chain=bool(group_futures),
            before_rebalancing=not bool(post_rebalancing),
        )
        title = 'Historical actual portfolio weights'
        return self._historical_portfolio_weights(df, title)

    def fig_historical_portfolio_weights_target(self, group_futures: bool=True):
        df = self.weights_target(
            aggregate_future_chain=bool(group_futures),
        )
        title='Historical target portfolio weights'
        return self._historical_portfolio_weights(df, title)

    def fig_historical_portfolio_weights_diff(
            self, group_futures: bool=True, post_rebalancing: bool=True
    ) -> go.Figure:
        df1 = self.weights_target(
            aggregate_future_chain=bool(group_futures),
        )
        df2 = self.weights_actual(
            aggregate_future_chain=bool(group_futures),
            before_rebalancing=not bool(post_rebalancing),
        )
        df = df1 - df2
        title = 'Historical target minus actual portfolio weights'
        return self._historical_portfolio_weights(df, title)


class TrackRecordComparison:
    def __init__(self, track_records: Sequence[TrackRecord]):
        self.track_records = self._parse_track_records(track_records)
        self._names = [tr.name for tr in self.track_records]
        if len(self._names) != len(set(self._names)):
            raise ValueError(
                'Duplicate track record names are now allowed:\n{}'
                ''.format(self._names)
            )

    def __getitem__(self, idx: Union[int, str]) -> TrackRecord:
        if isinstance(idx, str):
            idx = self._names.index(idx)
        return self.track_records[idx]

    def __repr__(self) -> str:
        return repr(self.track_records)

    @staticmethod
    def _parse_track_records(data) -> Sequence[TrackRecord]:
        track_records = list()
        if isinstance(data, TrackRecord):
            track_records.append(data)
        else:
            for item in data:
                if isinstance(item, str):
                    # Item is assumed to be a path to TrackRecord.
                    track_records.append(TrackRecord.load(item))
                elif isinstance(item, TrackRecord):
                    track_records.append(item)
                else:
                    raise ValueError("Unexpected 'data'.")
        return track_records

    def fig_net_liquidation_value(self, logy: bool=False) -> go.Figure:
        layout = go.Layout(
            title=go.layout.Title(
                text='Net liquidation value',
            ),
            yaxis=go.layout.YAxis(
                type='log' if logy else None,
                title=go.layout.yaxis.Title(text="$"),
            ),
        )
        data = list()
        for track_record in self.track_records:
            series = track_record.net_liquidation_value()
            scatter = go.Scatter(
                x=series.index,
                y=series.squeeze(),
                mode='lines',
                name=track_record.name,
                line=go.scatter.Line(width=2),
            )
            data.append(scatter)
        return go.Figure(data=data, layout=layout)



import numpy as np
from tradingenv.state import IState, cache
from tradingenv.events import Observer, IEvent
from datetime import datetime
import pytest
import gym


class EventSentiment(IEvent):
    def __init__(self, time: datetime, sentiment: float):
        self.time = time
        self.sentiment = sentiment


class TestStateWithoutSpace:
    def make_state(self):
        class ConcreteState(IState):
            def __init__(self):
                self.sentiment = 0.18

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

        return ConcreteState()

    def test_observed_events(self):
        state = self.make_state()
        assert state._observed_events == {'EventSentiment': 'process_EventSentiment'}

    def test_is_observer(self):
        assert issubclass(IState, Observer)

    def test_space_is_none_by_default(self):
        state = self.make_state()
        assert state.space is None

    def test_callable(self):
        state = self.make_state()
        assert state() is state

    def test_does_not_raise_if_no_observed_events(self):
        # Does not raise. Empty state with no features.
        IState()

    def test_callback(self):
        state = self.make_state()
        state.reset()
        assert state.sentiment == 0.18
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.sentiment == 0.42

    def test_not_raise_if_space_is_not_none_and_parse_is_not_implemented(self):
        class State(IState):
            def __init__(self):
                self.space = gym.spaces.Box(low=-1, high=1, shape=(1,), dtype=np.float64)
                self.sentiment = 0.18

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return np.array([self.sentiment])

        state = State()
        np.testing.assert_equal(
            actual=state(),
            desired=np.array([0.18])
        )

    def test_raise_if_state_does_not_belong_to_space(self):
        class State(IState):
            def __init__(self):
                self.space = gym.spaces.Box(low=-1, high=1, shape=(1,))
                self.sentiment = 2.  # greater than 1. will throw an exception

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return np.array([self.sentiment])

        state = State()
        with pytest.raises(ValueError):
            state()

    def test_not_raise_if_state_does_not_belong_to_space_and_False_verify(self):
        class State(IState):
            def __init__(self):
                self.space = gym.spaces.Box(low=-1, high=1, shape=(1,))
                self.sentiment = 2.  # greater than 1. will throw an exception

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return np.array([self.sentiment])

        state = State()
        state(verify=False)  # does not raise because verify=False


class TestStateHistory:
    def test_history_saved_by_default(self):
        class State(IState):
            def __init__(self):
                self.space = gym.spaces.Box(low=-1, high=1, shape=(1,), dtype=float)
                self.sentiment = 0.

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return np.array([self.sentiment])

        state = State()
        assert state.history == dict()
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.history == {datetime(2019, 1, 1): 0.42}

    @pytest.mark.skip(reason='Nothing bad happens if you dont implement parse')
    def test_parse_must_be_implemented_if_space_is_provided(self):
        """OpenAI-gym compatibility will fail if you don't implement parse,
        but an error will be raised probably when gym will check if state
        belongs to space. We could raise a clearer error message maybe?"""
        class State(IState):
            def __init__(self):
                self.space = gym.spaces.Box(low=-1, high=1, shape=(1,))
                self.sentiment = 0.

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

        state = State()
        with pytest.raises(NotImplementedError):
            EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])

    @pytest.mark.skip(reason='Nothing bad happens if you dont implement parse')
    def test_raise_if_space_is_not_none_and_parse_is_not_implemented(self):
        """OpenAI-gym compatibility will fail if you don't implement parse,
        but an error will be raised probably when gym will check if state
        belongs to space. We could raise a clearer error message maybe?"""
        class State(IState):
            def __init__(self):
                self.space = gym.spaces.Box(low=-1, high=1, shape=(1,))
                self.sentiment = 0.18

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

        state = State()
        with pytest.raises(NotImplementedError):
            state()

    def test_no_history_is_saved_if_parse_is_not_implemented(self):
        class State(IState):
            def __init__(self):
                self.sentiment = 0.

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

        state = State()
        assert state.history == dict()
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.history == dict()

    def test_save_with_space(self):
        class State(IState):
            def __init__(self):
                self.space = gym.spaces.Box(low=-1, high=1, shape=(1,), dtype=float)
                self.sentiment = 0.

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return np.array([self.sentiment])

        state = State()
        assert state.history == dict()
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.history == {
            datetime(2019, 1, 1): np.array([0.42]),
        }

    def test_save_without_space(self):
        class State(IState):
            def __init__(self):
                self.sentiment = 0.

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return np.array([self.sentiment])

        state = State()
        assert state.history == dict()
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.history == {
            datetime(2019, 1, 1): np.array([0.42]),
        }

    def test_parse_observe_raise_if_process_has_different_than_arg_event(self):
        class NewsSentiment(IState):
            def process_EventNews(self, event: EventSentiment, excess="a"):
                return 42

        with pytest.raises(TypeError):
            # NewsSentiment.process_EventNews must accept 'event' only.
            NewsSentiment()


class TestStateReset:
    def test_initialization_without_args_or_kwargs(self):
        class State(IState):
            def __init__(self):
                self.sentiment = 0.

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

        state = State()
        state.reset()
        assert state.last_update is None
        assert state.sentiment == 0.
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.last_update == datetime(2019, 1, 1)
        assert state.sentiment == 0.42
        state.reset()
        assert state.last_update is None
        assert state.sentiment == 0.

    def test_initialization_with_args_and_kwargs(self):
        class State(IState):
            def __init__(self, a, b, c='c'):
                self.a = a
                self.b = b
                self.c = c

            def process_EventSentiment(self, event: EventSentiment):
                self.a = event.sentiment
                self.b = event.sentiment
                self.c = event.sentiment

        state = State('a', b='b')
        state.reset()
        assert state.last_update is None
        assert state.a == 'a'
        assert state.b == 'b'
        assert state.c == 'c'
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.last_update == datetime(2019, 1, 1)
        assert state.a == 0.42
        assert state.b == 0.42
        assert state.c == 0.42
        state.reset()
        assert state.last_update is None
        assert state.a == 'a'
        assert state.b == 'b'
        assert state.c == 'c'

    def test_initialization_with_class_instance(self):
        class Sentiment:
            def __init__(self):
                self.value = 0.

        class State(IState):
            def __init__(self):
                self.sentiment = Sentiment()

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment.value = event.sentiment

        state = State()
        state.reset()
        assert state.last_update is None
        assert state.sentiment.value == 0.
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.last_update == datetime(2019, 1, 1)
        assert state.sentiment.value == 0.42
        state.reset()
        assert state.last_update is None
        assert state.sentiment.value == 0.

    def test_initialization_with_default_class_instance(self):
        class Sentiment:
            def __init__(self):
                self.value = 0.

        class State(IState):
            def __init__(self, sentiment):
                self.sentiment = sentiment

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment.value = event.sentiment

        state = State(Sentiment())
        state.reset()
        assert state.last_update is None
        assert state.sentiment.value == 0.
        EventSentiment(datetime(2019, 1, 1), 0.42).notify([state])
        assert state.last_update == datetime(2019, 1, 1)
        assert state.sentiment.value == 0.42
        state.reset()
        assert state.last_update is None

        # You are evil because you are using an object with mutable State and
        # reset will fail.
        assert state.sentiment.value != 0.

    def test_history_lifecycle_independent_from_resets(self):
        class State(IState):
            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return self.sentiment

        for _ in range(3):
            state = State()
            state.reset()
            assert state.history == dict()
            EventSentiment(datetime(2019, 1, 2), 42).notify([state])
            assert state.history == {datetime(2019, 1, 2): 42}

    def test_dont_save_if_false(self):
        class State(IState):
            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return self.sentiment

        for _ in range(3):
            state = State(save=False)
            state.reset()
            assert state.history == dict()
            EventSentiment(datetime(2019, 1, 2), 42).notify([state])
            assert state.history == dict()
            # assert state.history == {datetime(2019, 1, 2): 42}


class TestCache:
    def test_no_cache(self):
        class State(IState):
            def __init__(self):
                self.sentiment = 0.

            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return self.sentiment

        # First episode. Regardless the cache, everything is calculated from
        # scratch.
        state = State()
        EventSentiment(datetime(2019, 1, 1), 0).notify([state])
        EventSentiment(datetime(2019, 1, 2), 1).notify([state])
        EventSentiment(datetime(2019, 1, 3), 2).notify([state])
        assert state.history == {
            datetime(2019, 1, 1): 0,
            datetime(2019, 1, 2): 1,
            datetime(2019, 1, 3): 2,
        }

        # Second episode. No cache, so all updates are done from scratch again.
        state.reset()
        EventSentiment(datetime(2019, 1, 1), 1).notify([state])
        EventSentiment(datetime(2019, 1, 2), 2).notify([state])
        EventSentiment(datetime(2019, 1, 3), 3).notify([state])
        assert state.history == {
            datetime(2019, 1, 1): 1,
            datetime(2019, 1, 2): 2,
            datetime(2019, 1, 3): 3,
        }

    def test_cache_default_arguments(self):
        class State(IState):
            def __init__(self):
                self.sentiment = 0.

            @cache
            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return self.sentiment

        # First episode. Regardless the cache, everything is calculated from
        # scratch.
        state = State()
        EventSentiment(datetime(2019, 1, 1), 0).notify([state])
        EventSentiment(datetime(2019, 1, 2), 1).notify([state])
        EventSentiment(datetime(2019, 1, 3), 2).notify([state])
        assert state.history == {
            datetime(2019, 1, 1): 0,
            datetime(2019, 1, 2): 1,
            datetime(2019, 1, 3): 2,
        }

        # Second episode. Cache, so all updates are done are retrieved from
        # the first episode, despite being wrong in this case.
        # First callback is verified by default. So value != 0 would raise.
        state.reset()
        EventSentiment(datetime(2019, 1, 1), 0).notify([state])
        EventSentiment(datetime(2019, 1, 2), 2).notify([state])
        EventSentiment(datetime(2019, 1, 3), 3).notify([state])
        assert state.history == {
            datetime(2019, 1, 1): 0,  # this would be 1 without cache
            datetime(2019, 1, 2): 1,  # this would be 2 without cache
            datetime(2019, 1, 3): 2,  # this would be 3 without cache
        }

    def test_cache_arg_check_every(self):
        class State(IState):
            def __init__(self):
                self.sentiment = 0.

            @cache(check_every=2)
            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return self.sentiment

        # First episode. Regardless the cache, everything is calculated from
        # scratch.
        state = State()
        EventSentiment(datetime(2019, 1, 1), 0).notify([state])
        EventSentiment(datetime(2019, 1, 2), 1).notify([state])
        EventSentiment(datetime(2019, 1, 3), 2).notify([state])
        assert state.history == {
            datetime(2019, 1, 1): 0,
            datetime(2019, 1, 2): 1,
            datetime(2019, 1, 3): 2,
        }

        # Second episode. Cache with check_every=2, so first and third events
        # are checked.
        state.reset()
        # First event leads to the same state of checked first event, so no
        # error is raised.
        EventSentiment(datetime(2019, 1, 1), 0).notify([state])
        # The second event is different form the cached state (state=1), but
        # check_every=2 skips the check on this event.
        EventSentiment(datetime(2019, 1, 2), 2).notify([state])
        # The third event is different form the cached state (state=2).
        # check_every=2 so now there is basis_naive_tm1 check which raises.
        with pytest.raises(ValueError):
            EventSentiment(datetime(2019, 1, 3), 3).notify([state])

    def test_cache_arg_check_on_start(self):
        class State(IState):
            def __init__(self):
                self.sentiment = 0.

            @cache(check_on_start=True)
            def process_EventSentiment(self, event: EventSentiment):
                self.sentiment = event.sentiment

            def parse(self):
                return self.sentiment

        # First episode. Regardless the cache, everything is calculated from
        # scratch.
        state = State()
        EventSentiment(datetime(2019, 1, 1), 0).notify([state])
        EventSentiment(datetime(2019, 1, 2), 1).notify([state])
        EventSentiment(datetime(2019, 1, 3), 2).notify([state])
        assert state.history == {
            datetime(2019, 1, 1): 0,
            datetime(2019, 1, 2): 1,
            datetime(2019, 1, 3): 2,
        }

        # Second episode. Cache with check_every=2, so first and third events
        # are checked.
        state.reset()
        # First event leads to different state of checked first event, so
        # error is raised because check_on_start is True.
        with pytest.raises(ValueError):
            EventSentiment(datetime(2019, 1, 1), 1).notify([state])



# https://github.com/robertmartin8/PyPortfolioOpt
# https://quant.stackexchange.com/questions/tagged/portfolio-optimization?sort=votes&pageSize=15
# Fractional differentiation.
# Kelly (optimal f)
# RiskParity with custom definition of risk, eg Martin risk.
# Use OAS for robust covariance matrix estimation:
#    https://scikit-learn.org/stable/modules/generated/sklearn.covariance.OAS.html
"""Logic to wrap policy passed in TradingEnv.sample_episode(policy)."""
from tradingenv.spaces import PortfolioSpace
from abc import ABC, abstractmethod
import gym.spaces


class AbstractPolicy(ABC):
    # Class attributes are filled by make_policy()
    action_space: PortfolioSpace = None
    observation_space: gym.Space = None

    @abstractmethod
    def act(self, state):
        """Returns an action which belongs to the action observation_space, so
        action_space.contains(action) must return True."""

    def __repr__(self):
        return self.__class__.__name__


class RandomPolicy(AbstractPolicy):
    def act(self, state=None):
        return self.action_space.sample()


def make_policy(policy, action_space, observation_space) -> AbstractPolicy:
    if policy is None:
        policy = RandomPolicy()
    elif isinstance(policy, AbstractPolicy):
        pass
    else:
        try:
            from ray.rllib.policy.policy import Policy
            from ray.rllib.agents.trainer import Trainer
        except ModuleNotFoundError:
            pass
        else:
            from tradingenv.ray.walkforward.policy import RayAgent, RayPolicy
            if isinstance(policy, Trainer):
                policy = RayAgent(policy)
            elif isinstance(policy, Policy):
                policy = RayPolicy(policy)
    # Inject action and observation space.
    policy.action_space = action_space
    policy.observation_space = observation_space
    return policy



from tradingenv.metrics import to_pandas, PandasMetrics
from datetime import datetime
import pandas as pd
import numpy as np
import pytest
import unittest.mock
from pandas.core.generic import NDFrame


def test_to_pandas_default_args():
    df = pd.DataFrame()
    series = pd.Series(dtype=float)
    with pytest.raises(AttributeError):
        df.get_answer()
    with pytest.raises(AttributeError):
        series.get_answer()

    @to_pandas
    def get_answer(pandas_obj):
        return 42

    assert df.get_answer() == 42
    assert series.get_answer() == 42


def test_to_pandas_custom_target_class():
    df = pd.DataFrame()
    series = pd.Series(dtype=float)
    with pytest.raises(AttributeError):
        df.get_magic_number()
    with pytest.raises(AttributeError):
        series.get_magic_number()

    @to_pandas(target=pd.Series)
    def get_magic_number(pandas_obj):
        return 42

    with pytest.raises(AttributeError):
        df.get_magic_number()
    assert series.get_magic_number() == 42


class TestMetrics:
    def test_nr_days_with_non_business(self):
        index = pd.date_range("2017-12-31", "2018-12-31", freq="D")
        series = pd.Series(range(1, 1 + len(index)), index)
        assert series.nr_calendar_days() == 365
        assert series.nr_observations() == 366

    def test_nr_days_with_business(self):
        index = pd.date_range("2017-12-31", "2018-12-31", freq="B")
        series = pd.Series(range(1, 1 + len(index)), index)
        assert series.nr_calendar_days() == 364
        assert series.nr_observations() == 261

    def test_nr_days_with_unspecified_freq(self):
        index = pd.date_range("2017-12-31", "2018-12-31")
        series = pd.Series(range(1, 1 + len(index)), index)
        assert series.nr_calendar_days() == 365
        assert series.nr_observations() == 366

    def test_nr_years_with_non_business(self):
        index = pd.date_range("2017-12-31", "2018-12-31", freq="D")
        series = pd.Series(range(1, 1 + len(index)), index)
        assert series.nr_years() == 1

    def test_nr_years_with_business(self):
        index = pd.date_range("2017-12-31", "2018-12-31", freq="B")
        series = pd.Series(range(1, 1 + len(index)), index)
        assert series.nr_years() == 364 / 365

    def test_nr_years_with_unspecified_freq(self):
        index = pd.date_range("2017-12-31", "2018-12-31")
        series = pd.Series(range(1, 1 + len(index)), index)
        assert series.nr_years() == 1

    def test_level_returns_self(self):
        index = pd.date_range("2017-12-31", "2018-12-31")
        expected = pd.Series(range(1, 1 + len(index)), index)
        actual = expected.level()
        pd.testing.assert_series_equal(actual, expected)

    @unittest.mock.patch.object(pd.Series, "validate")
    def test_level_validates_self(self, mock_validate):
        index = pd.date_range("2017-12-31", "2018-12-31")
        series = pd.Series(range(1, 1 + len(index)), index)
        series.level()
        mock_validate.assert_called_once()

    def test_level_only_returns_last_observation_for_each_day(self):
        series = pd.Series(
            data={
                datetime(2019, 1, 1, 2): 1,
                datetime(2019, 1, 1, 20): 2,
                datetime(2019, 1, 3, 5, 15, 2): 3,
                datetime(2019, 1, 3, 5, 16, 1): 4,
                datetime(2019, 1, 3, 5, 16, 2): 5,
            }
        )
        actual = series.level()
        expected = pd.Series(data={datetime(2019, 1, 1): 2, datetime(2019, 1, 3): 5})
        actual.index = pd.to_datetime(actual.index)
        expected.index = pd.to_datetime(expected.index)
        pd.testing.assert_series_equal(actual, expected, check_index_type=False)

    def test_cagr(self):
        series = pd.Series([1, 2], [datetime(2017, 12, 31), datetime(2018, 12, 31)])
        assert series.cagr() == 1

    def test_log_returns(self):
        series = pd.Series([1, 2, 4], pd.date_range("2019-01-01", periods=3))
        expected = pd.Series(
            [np.log(2 / 1), np.log(4 / 2)], pd.date_range("2019-01-02", periods=2)
        )
        pd.testing.assert_series_equal(series.log_returns(), expected)

    def test_simple_returns(self):
        series = pd.Series([1, 2, 4], pd.date_range("2019-01-01", periods=3))
        expected = pd.Series(
            [2 / 1 - 1, 4 / 2 - 1], pd.date_range("2019-01-02", periods=2)
        )
        pd.testing.assert_series_equal(series.simple_returns(), expected)

    def test_simple_returns_with_nans(self):
        series = pd.Series([1, np.nan, 4], pd.date_range("2019-01-01", periods=3))
        with pytest.raises(ValueError):
            series.simple_returns()

    def test_volatility(self):
        series = pd.Series([1, 2, 4, 3, 6, 4], pd.date_range("2019-01-01", periods=6))
        expected = series.pct_change().iloc[1:].std() * np.sqrt(252)
        assert series.volatility() == expected

    def test_drawdown(self):
        series = pd.Series(
            [1, 2, 4, 3, 6, 4], pd.date_range("2019-01-01", periods=6), name="S&P 500"
        )
        expected = series / series.cummax() - 1
        pd.testing.assert_series_equal(series.drawdown(), expected)

    def test_max_drawdown(self):
        series = pd.Series(
            [1, 2, 4, 3, 6, 4], pd.date_range("2019-01-01", periods=6), name="S&P 500"
        )
        expected = 4 / 6 - 1
        assert series.max_drawdown() == expected

    def test_value_at_risk(self):
        # 102 prices -> 101 returns.
        series = pd.Series(range(1, 103), pd.date_range("2019-01-01", periods=102))
        assert series.value_at_risk(0) == 102 / 101 - 1
        assert series.value_at_risk(0.01) == 101 / 100 - 1
        assert series.value_at_risk(0.5) == (52 / 51) - 1
        assert series.value_at_risk(0.99) == 3 / 2 - 1
        assert series.value_at_risk(1) == 2 / 1 - 1

    def test_expected_shortfall(self):
        # 102 prices -> 101 returns.
        series = pd.Series(range(1, 103), pd.date_range("2019-01-01", periods=102))
        returns = series.pct_change().iloc[1:]
        assert series.expected_shortfall(0) == returns.iloc[100]
        assert series.expected_shortfall(0.01) == returns.iloc[99:].mean()
        assert series.expected_shortfall(0.5) == returns.iloc[50:].mean()
        assert series.expected_shortfall(0.99) == returns.iloc[1:].mean()
        assert series.expected_shortfall(1) == returns.mean()

    def test_downside_volatility(self):
        series = pd.Series(
            [1, 2, 1.8, 1.5, 1.2, 1.1],
            pd.date_range("2019-01-01", periods=6),
            name="S&P 500",
        )
        returns = series.pct_change().iloc[1:]
        expected = np.sqrt(252) * returns[returns < 0].std()
        assert series.downside_volatility() == expected

    def test_upside_volatility(self):
        series = pd.Series(
            [1, 2, 4, 3, 5, 7], pd.date_range("2019-01-01", periods=6), name="S&P 500"
        )
        returns = series.pct_change().iloc[1:]
        expected = np.sqrt(252) * returns[returns > 0].std()
        assert series.upside_volatility() == expected

    def test_cumulative_returns(self):
        series = pd.Series(
            [1, 1.2, 1.4, 1.3, 1.5, 1.7], pd.date_range("2019-01-01", periods=6)
        )
        actual = series.cumulative_return()
        expected = series / series.iloc[0] - 1
        assert actual.equals(expected)

    def test_cagr_equals_excess_return_over_zero(self):
        series = pd.Series(
            [1, 1.2, 1.4, 1.3, 1.5, 1.7], pd.date_range("2019-01-01", periods=6)
        )
        assert series.cagr() == series.excess_cagr(over=0.0)

    def test_validate_raises_if_missing_values(self):
        series = pd.Series([1, np.nan, 3], pd.date_range("2019-01-01", periods=3))
        with pytest.raises(ValueError):
            series.validate()

    def test_validate_raises_if_non_posive_values(self):
        series = pd.Series([1, -1, 3], pd.date_range("2019-01-01", periods=3))
        with pytest.raises(ValueError):
            series.validate()

    def test_validate_raises_if_duplicate_indices(self):
        series = pd.Series(
            [1, 2, 3],
            [datetime(2019, 1, 1), datetime(2019, 1, 2), datetime(2019, 1, 2)],
        )
        with pytest.raises(ValueError):
            series.validate()

    def test_validate_raises_if_not_DatetimeIndex(self):
        series = pd.Series([1, 2, 3])
        with pytest.raises(ValueError):
            series.validate()

    def test_validate_raises_if_index_has_nans(self):
        series = pd.Series(
            [1, 2, 3], [datetime(2019, 1, 1), np.nan, datetime(2019, 1, 2)]
        )
        with pytest.raises(ValueError):
            series.validate()

    def test_validate_raises_if_index_is_not_sorted(self):
        series = pd.Series(
            [1, 2, 3],
            [datetime(2019, 1, 2), datetime(2019, 1, 3), datetime(2019, 1, 1)],
        )
        with pytest.raises(ValueError):
            series.validate()


class TestTearsheet:
    def test_tearsheet(self):
        pass



"""It is harder to convert returns to levels than viceversa because the former
requires to know what was the base date."""
from typing import Callable, Type, Union
from pandas.core.generic import NDFrame
from functools import wraps
from collections import OrderedDict
from inspect import isclass
import numpy as np
import pandas as pd
from statsmodels.api import OLS
from statsmodels.regression.linear_model import RegressionResultsWrapper
from numbers import Number
import scipy.cluster.hierarchy as cluster

BDAYS = 252


def to_pandas(method: Callable = None, target: Type = NDFrame):
    """This decorator simply adds method to the signature of cls, which is
    NDFrame by default.

    Parameters
    ----------
    method : Callable
        A function definition whose first argument is an instance of cls.
    target : Type
        A class type. Input 'method' will be set as attribute of this 'target'
        class.


    Notes
    -----
    When optional arguments are passed to properties, then the first argument
    ('method' in this case) is None. This requires special handling, managed
    by the big if-else below. The implementation has been inspired by:
    https://stackoverflow.com/a/24617244/7677636

    Examples
    --------
    >>> df = pd.DataFrame()
    >>> series = pd.Series(dtype=float)
    >>> # df.method()  # raises
    >>> # series.method()  # raises
    ...
    >>> @to_pandas(target=pd.DataFrame)
    ... def method(frame: NDFrame):
    ...     return 'Hello!'
    >>> df.method()
    'Hello!'
    >>> # series.method()  # AttributeError: 'Series' object has no attribute 'method'
    ...
    >>> @to_pandas(target=pd.Series)
    ... def method(frame: NDFrame):
    ...     return 'Hello!'
    >>> df.method()
    'Hello!'
    >>> series.method()
    'Hello!'
    """
    if isclass(method):
        raise ValueError("to_pandas only supports keyword arguments.")
    optional_kwargs = method is None
    if optional_kwargs:

        @wraps(method)
        def decorated(method):
            setattr(target, method.__name__, method)
            return method

        return decorated
    else:
        setattr(target, method.__name__, method)
        return method


class PandasMetrics(NDFrame):
    """Define here methods to be added to pandas.Series AND pandas.DataFrame
    using the decorator to_pandas. All these methods assume that data are
    daily levels."""

    @to_pandas
    def validate(self):
        """Check that the data is a proper level dataset."""
        if np.any(np.any(np.isnan(self.values))):
            raise ValueError("Missing values are not allowed.")
        if np.any(np.any(self.values <= 0)):
            raise ValueError(
                "This method can only work with level data, so "
                "all values must be positive."
            )
        if self.index.has_duplicates:
            raise ValueError("Duplicate indices have been found")
        if not isinstance(self.index, pd.DatetimeIndex):
            raise ValueError("Input index must be a DatetimeIndex")
        if self.index.hasnans:
            raise ValueError("Input must not missing values in the index.")
        if not self.index.is_monotonic_increasing:
            raise ValueError("Input index must be monotonic increasing.")

    @to_pandas
    def level(self):
        """Returns the object itself, which is assumed to be a level. In case
        of duplicate dates, only the last timestamp for each day will be kept.
        """
        self.validate()
        if len(np.unique(self.index.date)) != len(self.index.date):
            self = self.groupby(by=self.index.date).last()
        return self

    @to_pandas
    def simple_returns(self):
        """
        Examples
        --------
        >>> prices = pd.Series([1, 2, 3], pd.date_range('2019-01-01', periods=3))
        >>> prices.simple_returns()
        2019-01-02    1.0
        2019-01-03    0.5
        Freq: D, dtype: float64
        """
        # https://github.com/pandas-dev/pandas/issues/6389
        level = self.level()
        # FutureWarning: The 'fill_method' and 'limit' keywords in
        # Series.pct_change are deprecated and will be removed in a future
        # version. Call ffill before calling pct_change instead.
        simple_returns = level.pct_change()  # fill_method=None
        return simple_returns.iloc[1:]

    @to_pandas
    def log_returns(self):
        """
        Examples
        --------
        >>> prices = pd.Series([1, 2, 3], pd.date_range('2019-01-01', periods=3))
        >>> prices.log_returns()
        2019-01-02    0.693147
        2019-01-03    0.405465
        Freq: D, dtype: float64
        """
        level = self.level()
        log_returns = np.log(level).diff()
        return log_returns.iloc[1:]

    @to_pandas
    def nr_calendar_days(self) -> int:
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02],
        ...     index=pd.date_range(start='2019-01-01', periods=5, freq='B'),
        ... )
        >>> prices.nr_calendar_days()
        6
        """
        return (self.last_valid_index() - self.first_valid_index()).days

    @to_pandas
    def nr_observations(self) -> int:
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02],
        ...     index=pd.date_range(start='2019-01-01', periods=5, freq='B'),
        ... )
        >>> prices.nr_observations()
        5
        """
        valid_obs = self.loc[self.first_valid_index() : self.last_valid_index()]
        return len(valid_obs)

    @to_pandas
    def nr_years(self) -> float:
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=np.linspace(1., 2., 366),
        ...     index=pd.date_range(start='2018-12-31', periods=366, freq='D'),
        ... )
        >>> prices.nr_years()
        1.0
        >>> prices.squeeze().nr_years()
        1.0
        """
        return self.nr_calendar_days() / 365

    @to_pandas
    def cagr(self):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=np.linspace(1., 1.5, 366),
        ...     index=pd.date_range(start='2018-12-31', periods=366, freq='D'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.cagr()
        S&P 500    0.5
        dtype: float64
        >>> prices.squeeze().cagr()
        0.5
        """
        level = self.level().bfill()
        return (level.iloc[-1] / level.iloc[0]) ** (1 / self.nr_years()) - 1

    @to_pandas
    def cumulative_return(self):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=np.linspace(1., 1.5, 10),
        ...     index=pd.date_range(start='2018-12-31', periods=10, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.cumulative_return()
                     S&P 500
        2018-12-31  0.000000
        2019-01-01  0.055556
        2019-01-02  0.111111
        2019-01-03  0.166667
        2019-01-04  0.222222
        2019-01-07  0.277778
        2019-01-08  0.333333
        2019-01-09  0.388889
        2019-01-10  0.444444
        2019-01-11  0.500000
        """
        level = self.level()
        level_start = level.loc[level.first_valid_index()]
        return level / level_start - 1

    @to_pandas
    def overall_return(self):
        return self.cumulative_return().iloc[-1]

    @to_pandas
    def volatility(self):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=np.linspace(1., 1.5, 10),
        ...     index=pd.date_range(start='2018-12-31', periods=10, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.volatility()
        S&P 500    0.092578
        dtype: float64
        >>> prices.squeeze().volatility()
        0.09257831...
        """
        return np.sqrt(BDAYS) * self.simple_returns().std()

    @to_pandas
    def drawdown(self):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03],
        ...     index=pd.date_range(start='2019-01-01', periods=6, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.drawdown()
                     S&P 500
        2019-01-01  0.000000
        2019-01-02  0.000000
        2019-01-03  0.000000
        2019-01-04 -0.038835
        2019-01-07 -0.009709
        2019-01-08  0.000000
        """
        level = self.level()
        return level / level.cummax() - 1

    @to_pandas
    def max_drawdown(self):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03],
        ...     index=pd.date_range(start='2019-01-01', periods=6, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.max_drawdown()
        S&P 500   -0.038835
        dtype: float64
        """
        return self.drawdown().min()

    @to_pandas
    def value_at_risk(self, quantile: float = 0.025):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03, 1., 0.98],
        ...     index=pd.date_range(start='2019-01-01', periods=8, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.value_at_risk(quantile=0.025)
        S&P 500   -0.037379
        Name: 0.025, dtype: float64
        """
        simple_returns = self.simple_returns()
        return simple_returns.quantile(quantile)

    @to_pandas
    def expected_shortfall(self, quantile: float = 0.025):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03, 1., 0.98],
        ...     index=pd.date_range(start='2019-01-01', periods=8, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.expected_shortfall(quantile=0.025)
        S&P 500   -0.038835
        dtype: float64
        """
        simple_returns = self.simple_returns()
        value_at_risk = simple_returns.quantile(quantile)
        tail = simple_returns[simple_returns <= value_at_risk]
        return tail.mean()

    @to_pandas
    def downside_volatility(self):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03, 1., 0.98],
        ...     index=pd.date_range(start='2019-01-01', periods=8, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.downside_volatility()
        S&P 500    0.149522
        dtype: float64
        """
        simple_returns = self.simple_returns()
        neg_returns = simple_returns[simple_returns < 0]
        return np.sqrt(BDAYS) * neg_returns.std()

    @to_pandas
    def upside_volatility(self):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03, 1., 0.98],
        ...     index=pd.date_range(start='2019-01-01', periods=8, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.upside_volatility()
        S&P 500    0.154643
        dtype: float64
        """
        simple_returns = self.simple_returns()
        pos_returns = simple_returns[simple_returns > 0]
        return np.sqrt(BDAYS) * pos_returns.std()

    @to_pandas
    def martin_risk(self):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03, 1., 0.98],
        ...     index=pd.date_range(start='2019-01-01', periods=8, freq='B'),
        ...     columns=['S&P 500'],
        ... )
        >>> prices.martin_risk()
        S&P 500    0.024513
        dtype: float64
        """
        # http://www.tangotools.com/ui/ui.htm
        # https://en.wikipedia.org/wiki/Ulcer_index
        return np.sqrt(self.drawdown().pow(2).mean())

    @to_pandas
    def tracking_error(self, other: NDFrame):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data={
        ...         'S&P 500': [1., 1.01, 1.03, 0.99, 1.02, 1.03, 1., 0.98],
        ...         'SPY': [1., 1.011, 1.0298, 0.991, 1.0208, 1.031, 1., 0.979],
        ...     },
        ...     index=pd.date_range(start='2019-01-01', periods=8, freq='B'),
        ... )
        >>> prices['S&P 500'].tracking_error(prices['SPY'])
        0.01536269...
        """
        excess_returns = self.excess_returns(other)
        return np.sqrt(BDAYS) * excess_returns.std()

    @to_pandas
    def _parse_rate(self, rate: Union[NDFrame, float]) -> float:
        if isinstance(rate, NDFrame):
            rate = rate.squeeze().cagr()
        return rate

    @to_pandas
    def excess_returns(self, other: NDFrame):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data={
        ...         'S&P 500': (1, 1.15, 1.18, 1.12),
        ...         'RiskFreeRate': (1, 1.02, 1.03, 1.04),
        ...     },
        ...     index=pd.date_range(start='2019-01-01', periods=4, freq='D'),
        ...     columns=['S&P 500', 'RiskFreeRate'],
        ... )
        >>> prices['S&P 500'].excess_returns(prices['RiskFreeRate'])
        2019-01-02    0.130000
        2019-01-03    0.016283
        2019-01-04   -0.060556
        Freq: D, dtype: float64
        """
        self_returns = self.simple_returns()
        other_returns = other.simple_returns().squeeze().reindex(self_returns.index)
        excess_returns = self_returns.subtract(other_returns, axis=0)
        return excess_returns

    @to_pandas
    def excess_cagr(self, over: Union[NDFrame, float] = 0.0):
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data={
        ...         'S&P 500': np.linspace(1, 1.15, 366),
        ...         'RiskFreeRate': np.linspace(1, 1.02, 366),
        ...     },
        ...     index=pd.date_range(start='2019-01-01', periods=366, freq='D'),
        ...     columns=['S&P 500', 'RiskFreeRate'],
        ... )
        >>> prices.excess_cagr(over=0.) == prices.cagr()
        S&P 500         True
        RiskFreeRate    True
        dtype: bool
        >>> prices.excess_cagr(over=0.02).round(6)
        S&P 500         0.13
        RiskFreeRate    0.00
        dtype: float64
        """
        return self.cagr() - self._parse_rate(over)

    @to_pandas
    def sharpe_ratio(self, risk_free: Union[NDFrame, float] = 0.0) -> float:
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data={
        ...         'S&P 500': [1., 1.01, 1.03, 1.03, 1., 1.01],
        ...         'RiskFreeRate': [1., 1.001, 1005, 1.006, 1.009, 1.01],
        ...     },
        ...     index=pd.date_range(start='2019-01-01', periods=6, freq='B'),
        ...     columns=['S&P 500', 'RiskFreeRate'],
        ... )
        >>> prices.sharpe_ratio(risk_free=prices['RiskFreeRate'])
        S&P 500         0.0
        RiskFreeRate    0.0
        dtype: float64
        >>> prices.sharpe_ratio(risk_free=prices['RiskFreeRate'].cagr())
        S&P 500         0.0
        RiskFreeRate    0.0
        dtype: float64
        >>> prices = pd.DataFrame(
        ...     data={
        ...         'S&P 500': [1., 1.01, 1.03, 1.03, 1., 1.03],
        ...         'RiskFreeRate': [1., 1.001, 1005, 1.007, 1.009, 1.01],
        ...     },
        ...     index=pd.date_range(start='2019-01-01', periods=6, freq='B'),
        ...     columns=['S&P 500', 'RiskFreeRate'],
        ... )
        >>> prices['S&P 500'].sharpe_ratio(risk_free=prices['RiskFreeRate'])
        8.31680812940463
        """
        return self.excess_cagr(risk_free) / self.volatility()

    @to_pandas
    def sortino_ratio(self, risk_free: Union[NDFrame, float] = 0.0) -> float:
        """
        Examples
        --------
        >>> prices = pd.DataFrame(
        ...     data={
        ...         'S&P 500': [1., 1.01, 1.02, 1.05, 1.03, 1.02],
        ...         'RiskFreeRate': [1., 1.001, 1005, 1.007, 1.009, 1.01],
        ...     },
        ...     index=pd.date_range(start='2019-01-01', periods=6, freq='B'),
        ...     columns=['S&P 500', 'RiskFreeRate'],
        ... )
        >>> prices['S&P 500'].sortino_ratio(risk_free=prices['RiskFreeRate'])
        10.7621879372323
        >>> risk_free_cagr = prices['RiskFreeRate'].cagr()
        >>> risk_free_cagr
        0.6800754114925203
        >>> prices['S&P 500'].sortino_ratio(risk_free_cagr)
        10.7621879372323
        """
        return self.excess_cagr(risk_free) / self.downside_volatility()

    @to_pandas
    def information_ratio(self, benchmark: NDFrame) -> float:
        return self.excess_cagr(benchmark) / self.tracking_error(benchmark)

    @to_pandas
    def calmar_ratio(self, risk_free: Union[NDFrame, float] = 0.0) -> float:
        return self.excess_cagr(risk_free) / -self.max_drawdown()

    @to_pandas
    def martin_ratio(self, risk_free: Union[NDFrame, float] = 0.0) -> float:
        # http://www.tangotools.com/ui/ui.htm
        # https://en.wikipedia.org/wiki/Ulcer_index
        # https://www.keyquant.com/Download/GetFile?Filename=%5CPublications%5CKeyQuant_WhitePaper_APT_Part1.pdf
        return self.excess_cagr(risk_free) / self.martin_risk()

    @to_pandas
    def omega_ratio(self, risk_free: Union[NDFrame, float] = 0.0, threshold: float = None):
        # Code from aric.
        # I'm not convinced with data quality and wouldn't trust the output.
        #if isinstance(risk_free, Number):
        #    risk_free = self.make_series_from_cagr(risk_free, "RiskFree")
        risk_free = self._parse_rate(risk_free)
        if threshold is None:
            threshold = (1 + risk_free) ** (1 / BDAYS) - 1

        def lpm(returns, threshold, order):
            # This method returns a lower partial moment of the returns
            # Create an array he same length as returns containing the minimum return threshold
            #threshold_array = np.empty(len(returns))
            #threshold_array.fill(threshold)
            # Calculate the difference between the threshold and the returns
            diff = threshold - returns
            # Set the minimum of each to 0
            diff = diff.clip(lower=0)
            # Return the sum of the different to the power of order
            return np.sum(diff ** order, axis=0) / len(returns) * np.sqrt(BDAYS)

        excess_cagr = self.excess_cagr(risk_free)
        return excess_cagr / lpm(self.simple_returns(), threshold, 1)

    @to_pandas
    def capm(
        self, benchmark: NDFrame, risk_free: Union[NDFrame, float] = 0.0
    ) -> RegressionResultsWrapper:
        """Run a regression for each asset (column) of self."""
        if isinstance(risk_free, Number):
            risk_free = self.make_series_from_cagr(risk_free, "RiskFree")
        benchmark = benchmark.squeeze()

        x = benchmark.excess_returns(risk_free)
        x = x.to_frame('beta')
        x["alpha"] = 1.0
        y = self.squeeze().excess_returns(risk_free)
        common_idx = x.index.intersection(y.index)
        linear_regression = OLS(
            endog=y.loc[common_idx],
            exog=x.loc[common_idx],
            hasconst=True,
        )
        return linear_regression.fit()

    @to_pandas
    def make_series_from_cagr(self, cagr: float, name: str = None) -> pd.Series:
        """
        Parameters
        ----------
        cagr : float
            Annualized compounded growth rate of the series that you want
            to get.
        name : str
            Name of the returned pd.Series.

        Returns
        -------
        Returns a series with the same index and values corresponding to
        a level series growing at the specified rate.

        Examples
        --------
        >>> series = pd.Series(index=pd.date_range('2019-01-01', periods=BDAYS+1), dtype=float)
        >>> prices = series.make_series_from_cagr(0.01)
        >>> prices.head()
        2019-01-01    1.000000
        2019-01-02    1.000039
        2019-01-03    1.000079
        2019-01-04    1.000118
        2019-01-05    1.000158
        Freq: D, dtype: float64
        >>> prices.tail()
        2019-09-06    1.00984
        2019-09-07    1.00988
        2019-09-08    1.00992
        2019-09-09    1.00996
        2019-09-10    1.01000
        Freq: D, dtype: float64
        """
        rate_daily = (1 + cagr) ** (1 / BDAYS)
        series = pd.Series(data=rate_daily, index=self.index, name=name)
        series.iloc[0] = 1.0
        return series.cumprod()

    @to_pandas
    def alpha(self, benchmark: NDFrame, risk_free: Union[NDFrame, float] = 0.0):
        """Returns alpha coefficient of CAPM regression."""
        capm = self.capm(benchmark, risk_free)
        return (1 + capm.params["alpha"]) ** BDAYS - 1

    @to_pandas
    def beta(self, benchmark: NDFrame, risk_free: Union[NDFrame, float] = 0.0):
        """Returns beta coefficient of CAPM regression."""
        capm = self.capm(benchmark, risk_free)
        return capm.params['beta']

    @to_pandas
    def _intersect_index(self, other: NDFrame):
        idx = self.index.intersection(other.index)
        return other.loc[idx]

    @to_pandas
    def tearsheet(
        self,
        benchmark: NDFrame = None,
        risk_free: Union[NDFrame, float] = 0.0,
        weights: NDFrame = None,
        prices: NDFrame = None,
    ):
        """
        Calculate a tearsheet using daily finacial time series.

        Parameters
        ----------
        benchmark : NDFrame
            A NDFrame representing the levels series that you want to use to
            benchmark your investment strategy. This argument is mostly used
            when rendering the environment or calculating tearsheets.
        risk_free : Union[NDFrame, float]
            Risk free rate of the market (e.g. T-Bills, EURIBOR) expressed as
            a level pandas Series or DataFrame.
        weights : NDFrame
            Historical weights weights of the strategy. These can be found in
            TrackRecord.target_weights.
        prices : pd.DataFrame
            A pandas.DataFrame whose columns are asset _names, index are
            timestamps and values are prices.


        Returns
        -------
        A pd.DataFrame with several return, risk and risk-adjusted return
        metrics.

        Examples
        -------
        >>> returns = pd.DataFrame(
        ...     data=np.random.normal(0, 0.01, (1000, 3)),
        ...     index=pd.date_range('2010', freq='D', periods=1000),
        ...     columns=['Strategy', 'S&P 500', 'T-Notes']
        ... )
        >>> level = (1 + returns).cumprod()
        >>> tearsheet = level['Strategy'].tearsheet(
        ...     benchmark=level['S&P 500'],
        ...     risk_free=level['T-Notes'],
        ... )
        """
        if isinstance(risk_free, Number):
            risk_free = self.make_series_from_cagr(risk_free, "RiskFree")

        # TODO: test tearsheet
        # Ensure that we analyse data over the same time span (daily data)
        index = self.index
        for data in [benchmark, risk_free, weights, prices]:
            if data is not None:
                index = index.intersection(data.index)
        self = self.loc[index]
        risk_free = risk_free.loc[index]

        # Construct the tearsheet.
        tearsheet = OrderedDict()
        tearsheet[("Context", "From")] = self.first_valid_index().date()
        tearsheet[("Context", "To")] = self.last_valid_index().date()
        tearsheet[("Context", "Years")] = self.nr_years()
        tearsheet[("Context", "Observations")] = self.nr_observations()
        tearsheet[("Context", "Risk-free asset")] = str(risk_free.squeeze().name)
        tearsheet[("Context", "Risk-free CAGR")] = self._parse_rate(risk_free)
        tearsheet[("Return", "CAGR")] = self.cagr()
        tearsheet[("Return", "CAGR over cash")] = self.excess_cagr(risk_free)
        tearsheet[("Return", "Overall return")] = self.overall_return()
        tearsheet[("Risk", "Volatility")] = self.volatility()
        tearsheet[("Risk", "Downside volatility")] = self.downside_volatility()
        tearsheet[("Risk", "Upside volatility")] = self.upside_volatility()
        tearsheet[("Risk", "Max drawdown")] = self.max_drawdown()
        tearsheet[("Risk", "Martin risk")] = self.martin_risk()
        tearsheet[("Risk", "VaR 5%")] = self.value_at_risk(0.05)
        tearsheet[("Risk", "VaR 2%")] = self.value_at_risk(0.02)
        tearsheet[("Risk", "Expected shortfall 5%")] = self.expected_shortfall(0.05)
        tearsheet[("Risk", "Expected shortfall 2%")] = self.expected_shortfall(0.02)
        tearsheet[("Risk-adjusted return", "Sharpe ratio")] = self.sharpe_ratio(
            risk_free
        )
        tearsheet[("Risk-adjusted return", "Sortino ratio")] = self.sortino_ratio(
            risk_free
        )
        tearsheet[("Risk-adjusted return", "Calmar ratio")] = self.calmar_ratio(
            risk_free
        )
        tearsheet[("Risk-adjusted return", "Martin ratio")] = self.martin_ratio(
            risk_free
        )
        tearsheet[("Risk-adjusted return", "Omega ratio")] = self.omega_ratio(
            risk_free
        )
        if benchmark is not None:
            benchmark = benchmark.loc[index]
            tearsheet[("Outperformance", "Benchmark id")] = str(benchmark.squeeze().name)
            tearsheet[("Outperformance", "CAGR over benchmark")] = self.excess_cagr(
                benchmark
            )
            tearsheet[("Outperformance", "Information ratio")] = self.information_ratio(
                benchmark
            )
            tearsheet[("Outperformance", "CAPM Alpha")] = self.alpha(
                benchmark, risk_free
            )
            tearsheet[("Outperformance", "CAPM Beta")] = self.beta(benchmark, risk_free)
            tearsheet[("Outperformance", "Correlation")] = self.simple_returns().squeeze().corr(benchmark.simple_returns().squeeze())
        if weights is not None:
            weights = weights.loc[index]
            leverage = weights.sum(1)
            turnover = weights.diff().abs().sum(1)
            tearsheet[("Weights", "Leverage mean")] = leverage.mean()
            tearsheet[("Weights", "Turnover daily")] = turnover.mean()
            tearsheet[("Weights", "Turnover annual")] = BDAYS * turnover.mean()
            for asset_name, weight_avg in weights.mean().items():
                tearsheet[("Weights", str(asset_name))] = weight_avg
        if prices is not None:
            prices = prices.loc[index]
            for name, asset in prices.items():
                if asset.std() > 0:  # avoid stuff like cash
                    tearsheet[("Markets", "CAGR " + str(name))] = asset.cagr()
        return pd.DataFrame(self.__class__(tearsheet).T)


class SeriesMetrics(NDFrame):
    """Define here methods to be added to pandas.Series. All these methods
    assume that data are daily levels."""

    @to_pandas(target=pd.Series)
    def drawdown_periods(self):
        """
        Examples
        --------
        >>> prices = pd.Series(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03, 1., 0.98],
        ...     index=pd.date_range(start='2019-01-01', periods=8, freq='B'),
        ...     name='S&P 500',
        ... )
        >>> prices.drawdown_periods()
           Max drawdown       From     Trough         To  Nr days
        1     -0.048544 2019-01-09 2019-01-10 2019-01-10        2
        0     -0.038835 2019-01-04 2019-01-04 2019-01-07        2
        """
        drawdown = self.drawdown()
        group_ids = (drawdown == 0).cumsum()
        group_ids = group_ids[drawdown != 0]
        groups = drawdown.groupby(group_ids)
        drawdown_periods = pd.DataFrame(
            {
                "Max drawdown": groups.min(),
                "From": groups.aggregate(lambda series: series.index.min()),
                "Trough": groups.idxmin(),
                "To": groups.aggregate(lambda series: series.index.max()),
                "Nr days": groups.count(),
            }
        )
        drawdown_periods.reset_index(inplace=True, drop=True)
        return drawdown_periods.sort_values(by="Max drawdown")

    @to_pandas(target=pd.Series)
    def linear_regression(self, reset_index: bool=False):
        """
        Examples
        -------
        >>> prices = pd.Series(
        ...     data=[1., 1.01, 1.03, 0.99, 1.02, 1.03, 1., 0.98],
        ...     index=pd.date_range(start='2019-01-01', periods=8, freq='B'),
        ...     name='S&P 500',
        ... )
        >>> trendline = prices.linear_regression(reset_index=True)
        >>> trendline['trendline']
        2019-01-01    1.014167
        2019-01-02    1.012262
        2019-01-03    1.010357
        2019-01-04    1.008452
        2019-01-07    1.006548
        2019-01-08    1.004643
        2019-01-09    1.002738
        2019-01-10    1.000833
        Freq: B, dtype: float64
        >>> trendline['residuals']
        2019-01-01   -0.014167
        2019-01-02   -0.002262
        2019-01-03    0.019643
        2019-01-04   -0.018452
        2019-01-07    0.013452
        2019-01-08    0.025357
        2019-01-09   -0.002738
        2019-01-10   -0.020833
        Freq: B, dtype: float64
        >>> trendline['std']  # doctest: +ELLIPSIS
        0.01771850...
        """
        y = self
        if reset_index:
            x = pd.DataFrame(np.arange(len(self)), index=self.index)
        else:
            x = self.index.to_frame()
        x['constant'] = 1.
        ols = OLS(endog=y, exog=x, hasconst=True)
        ols_results = ols.fit()
        return {
            "model": ols_results,
            "level": self,
            "trendline": ols_results.predict(x),
            "residuals": ols_results.resid,
            "std": ols_results.resid.std(),
        }


class DataFrameMetrics(NDFrame):
    """Define here methods to be added to pandas.Series. All these methods
    assume that data are daily levels."""

    @to_pandas(target=pd.DataFrame)
    def covariance_matrix(self):
        """Returns covariance matrix of simple returns."""
        returns = self.simple_returns()
        return returns.cov()

    @to_pandas(target=pd.DataFrame)
    def correlation_matrix(self, clustering: bool = False):
        """Returns correlation matrix. If clustering=True, the correlation
        matrix will be sorted accoring to the clusters and clusters will be
        returned."""
        returns = self.simple_returns()
        corr_mat = returns.corr()
        if clustering:
            distances = cluster.distance.pdist(corr_mat.values)
            linkages = cluster.linkage(distances, method="complete")
            ind = cluster.fcluster(linkages, 0.5 * distances.max(), "distance")
            columns = [corr_mat.columns[i] for i in np.argsort(ind)]
            corr_mat = corr_mat.reindex(columns, axis=0).reindex(columns, axis=1)

            # Formatting of clusters.
            clusters = dict()
            for i in set(ind):
                mask = ind == i
                clusters[i] = corr_mat.columns[mask].to_list()
            return corr_mat, clusters
        else:
            return corr_mat



"""Classes to store the state of the exchange, used to check buy and sell
prices. Exchange is an aggregation of LimitOrderBook."""
from tradingenv.events import Observer, EventNBBO, EventContractDiscontinued
from tradingenv.contracts import AbstractContract
from datetime import datetime
from typing import Sequence, Dict, Union
from collections import defaultdict
import numpy as np
import pandas as pd


class LimitOrderBook:
    """Limit order book of a single asset."""

    def __init__(
        self,
        bid_price: float = np.nan,
        ask_price: float = np.nan,
        bid_size: float = np.nan,
        ask_size: float = np.nan,
        time: datetime = None,
    ):
        """
        Parameters
        ----------
        bid_price : float
            Latest on-screen bid transaction_price.
        ask_price : float
            Latest on-screen ask transaction_price.
        bid_size : float
            Latest on-screen bid size.
        ask_size : float
            Latest on-screen ask size.

        Examples
        --------
        >>> from tradingenv.exchange import LimitOrderBook
        >>> lob = LimitOrderBook(
        ...     bid_price=99.9,
        ...     ask_price=100.1,
        ...     bid_size=100,
        ...     ask_size=200,
        ... )
        >>> lob
        99.9 : 100.1
        """
        # Inputs.
        self.bid_price = bid_price
        self.ask_price = ask_price
        self.bid_size = bid_size
        self.ask_size = ask_size
        self.time = time

        # More attributes.
        self.history = defaultdict(list)
        self.is_alive = True

    def __repr__(self) -> str:
        """Class string representation."""
        return "{bid_price} : {ask_price}" "".format(
            bid_price=self.bid_price,
            ask_price=self.ask_price,
        )

    @property
    def spread(self) -> float:
        """Returns bid-ask spread."""
        return self.ask_price - self.bid_price

    @property
    def mid_price(self) -> float:
        """Returns the middle transaction_price between the bid and the ask."""
        return (self.ask_price + self.bid_price) / 2

    def liq_price(self, quantity: float) -> float:
        """Returns the execution price of selling 'quantity' now."""
        return self.acq_price(-quantity)

    def acq_price(self, quantity: float) -> float:
        """Returns the execution price of buying 'quantity' now."""
        if quantity < 0:
            return self.bid_price
        elif quantity > 0:
            return self.ask_price
        elif quantity == 0:
            return self.mid_price
        else:
            raise ValueError("Unexpected sign: {}".format(quantity))

    def update(self, event: EventNBBO):
        # NOTE: this Exchange is responsible to call this method rather than
        # making LOB an Observe for lecagy reasons.
        self.bid_price = event.bid_price
        self.ask_price = event.ask_price
        self.bid_size = event.bid_size
        self.ask_size = event.ask_size
        self.time = event.time

        # Save historical data.
        self.history["time"].append(self.time)
        self.history["bid_price"].append(self.bid_price)
        self.history["ask_price"].append(self.ask_price)
        self.history["mid_price"].append(self.mid_price)
        self.history["bid_size"].append(self.bid_size)
        self.history["ask_size"].append(self.ask_size)

    def terminate(self, event: EventContractDiscontinued):
        """Reset and stop quotes in the lob but keep history."""
        history = self.history
        self.__init__()
        self.history = history
        self.time = event.time
        self.is_alive = False

    def to_frame(self, what: str='mid_price') -> pd.Series:
        return pd.Series(
            data=self.history[what],
            index=self.history['time'],
            name=what,
        )


class Exchange(Observer):
    """An aggregation of limit order books.

    Attributes
    ----------
    _books : Dict[AbstractContract, LimitOrderBook]
        A dictionary mapping contract IDs to the corresponding instance of
        LimitOrderBook.
    """

    def __init__(self):
        """
        Examples
        --------
        >>> from tradingenv.exchange import Exchange
        >>> from tradingenv.events import EventNBBO
        >>> from tradingenv.contracts import ETF
        >>> exchange = Exchange()
        >>>
        >>> # Assume that there is a NBBO update for S&P 500 and NASDAQ.
        >>> nbbo_sp500 = EventNBBO(
        ...     time=datetime.now(),
        ...     bid_price=9,
        ...     ask_price=11,
        ...     bid_size=100,
        ...     ask_size=200,
        ...     contract=ETF('SPY'),
        ... )
        >>> nbbo_nasdaq = EventNBBO(
        ...     time=datetime.now(),
        ...     bid_price=99,
        ...     ask_price=101,
        ...     bid_size=10,
        ...     ask_size=20,
        ...     contract=ETF('IEF'),
        ... )
        >>> exchange.process_EventNBBO(nbbo_sp500)
        >>> exchange.process_EventNBBO(nbbo_nasdaq)
        >>> exchange
        {ETF(SPY): '9 : 11', ETF(IEF): '99 : 101'}
        """
        self._books = defaultdict(LimitOrderBook)
        self.last_update = None

    def __repr__(self) -> str:
        """Class string representation."""
        return str(
            {contract_id: repr(book)
             for contract_id, book in self._books.items()
             if not np.isnan(book.mid_price)
             }
        )

    def __len__(self) -> int:
        """Returns number of limit order books in the exchange."""
        return len(self._books)

    def __getitem__(self, key: Union[str, AbstractContract]) -> LimitOrderBook:
        """Returns the limit order book associated with contract_id."""
        if isinstance(key, AbstractContract):
            key = key.static_hashing()  # TODO: Test
        return self._books[key]

    def bid_prices(self, keys: Sequence[AbstractContract]) -> np.array:
        """Return a np.array with mid prices of the provided contract IDs."""
        return np.array([self[key].bid_price for key in keys])

    def ask_prices(self, keys: Sequence[AbstractContract]) -> np.array:
        """Return a np.array with mid prices of the provided contract IDs."""
        return np.array([self[key].ask_price for key in keys])

    def mid_prices(self, keys: Sequence[AbstractContract]) -> np.array:
        """Return a np.array with mid prices of the provided contract IDs."""
        return np.array([self[key].mid_price for key in keys])

    def acq_prices(
        self, keys: Sequence[AbstractContract], signs: np.ndarray
    ) -> np.ndarray:
        return np.array([self[key].acq_price(sign) for key, sign in zip(keys, signs)])

    def liq_prices(
        self, keys: Sequence[AbstractContract], signs: np.ndarray
    ) -> np.ndarray:
        return self.acq_prices(keys, -signs)

    def spreads(self, keys: Sequence[AbstractContract]) -> np.array:
        """Return a np.array with spreads of the provided contract IDs."""
        return np.array([self[key].spread for key in keys])

    def process_EventNBBO(self, event: EventNBBO):
        """
        Process an EventNBBO, representing an update of market prices and
        sizes in the limit order book.

        Parameters
        ----------
        event : EventNBBO
            An event instance storing prices, sizes and contract_id to be
            updated.
        """
        book = self[event.contract]
        if book.is_alive:
            book.update(event)
        self.last_update = event.time

    def process_EventContractDiscontinued(self, event: EventContractDiscontinued):
        self[event.contract].terminate(event)

    def to_frame(self, contracts: Sequence[AbstractContract], join: str = 'outer'):
        # TODO: Test.
        # An alternative implementation, but not faster due to slow .from_dict
        #contracts = result.policy._returns.columns
        #raw = dict()
        #for contract in contracts:
        #    history = result.env.exchange[contract].history
        #    raw[contract] = dict(zip(history['time'], history['mid_price']))
        #pd.DataFrame.from_dict(raw).sort_index()
        data = list()
        for contract in contracts:
            history = self[contract].history
            series = pd.Series(
                data=history['mid_price'],
                index=history['time'],
                name=contract,
                dtype=np.float32,
            )
            data.append(series)
        return pd.concat(data, axis='columns', join=join, sort=True, copy=False)



from tradingenv.spaces import (
    Set,
    Float,
    PortfolioSpace,
    DiscretePortfolio,
    BoxPortfolio,
)
from tradingenv.broker.rebalancing import Weights
from tradingenv.broker.rebalancing import Rebalancing
from tradingenv.contracts import Cash, ETF, ES
from tradingenv.contracts import Future, FutureChain
from tradingenv.contracts import FutureChain, ES
from gym.spaces import Space, Discrete, Box
from datetime import datetime
import pytest
import numpy as np
import gym


try:
    from ray.rllib.models.extra_spaces import Simplex
    RAY_NOT_INSTALLED = False
except ModuleNotFoundError:
    RAY_NOT_INSTALLED = True


class TestSet:
    def test_parent_is_gym_space(self):
        assert issubclass(Set, gym.Space)

    def test_initialisation(self):
        space = Set(-1, 1)
        assert space.items == (-1, 1)

    def test_sample(self):
        space = Set(-1, 1)
        assert space.sample() in (-1, 1)

    def test_sample_distribution_is_uniform(self):
        space = Set(-1, 1)
        samples = np.array([space.sample() for _ in range(10000)])
        assert -0.05 < samples.mean() < 0.05

    def test_sample_mixed_dtypes(self):
        space = Set('a', 1)
        samples = [space.sample() for _ in range(20)]
        assert 'a' in samples
        assert 1 in samples
        assert '1' not in samples

    def test_contains(self):
        space = Set(-1, 1)
        assert space.contains(-1)
        assert space.contains(1)
        assert not space.contains(0)


class TestFloat:
    def test_parent_is_gym_space(self):
        assert issubclass(Float, gym.Space)

    def test_initialisation(self):
        low = -1
        high = 1
        space = Float(low, high)
        assert space.low == low
        assert space.high == high

    def test_default_initialization(self):
        space = Float()
        assert space.low == -1e16
        assert space.high == 1e16

    def test_sample_distribution_is_unbiased(self):
        low = -1
        high = 1
        space = Float(low=low, high=high)
        samples = np.array([space.sample() for _ in range(10000)])
        assert -0.05 < samples.mean() < 0.05

    def test_sample_is_within_bounds(self):
        low = -1
        high = 1
        space = Float(low=low, high=high)
        samples = [space.sample() for _ in range(10000)]
        assert min(samples) > -1
        assert max(samples) < 1

    def test_contains(self):
        low = -1
        high = 1
        space = Float(low=low, high=high)
        assert not space.contains(-1.01)
        assert space.contains(-1)
        assert space.contains(1)
        assert not space.contains(1.01)

    def test_sample_from_default_bounds(self):
        # Make sure that the following error is not raised:
        #  OverflowError: Range exceeds valid bounds
        space = Float()
        space.sample()


class TestPortfolioSpace:
    def test_make_allocation_is_abstract_method(self):
        space = PortfolioSpace([Cash(), ETF("IEF")])
        with pytest.raises(NotImplementedError):
            space._make_allocation(0)

    def test_default_initialization(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.ones((len(self.contracts),))

        space = MockPortfolioSpace([Cash(), ETF("IEF")])
        assert space.contracts == [Cash(), ETF("IEF")]
        assert space._as_weights is True
        assert space._fractional is True
        assert space._margin == 0
        assert space.base_currency == Cash()

    def test_custom_initialization(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.ones((len(self.contracts),))

        space = MockPortfolioSpace([Cash('EUR'), ETF("IEF")], True, False, 0.02)
        assert space.contracts == [Cash('EUR'), ETF("IEF")]
        assert space._as_weights is True
        assert space._fractional is False
        assert space._margin == 0.02
        assert space.base_currency == Cash('EUR')

    def test_raises_if_contracts_are_not_all_abstract_contracts(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.ones((len(self.underlyings),))

        match = "'contracts' must be a sequence of AbstractContract instances."
        with pytest.raises(TypeError, match=match):
            MockPortfolioSpace(["SPY", ETF("IEF")])

    def test_raises_if_all_contracts_are_not_unique(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.ones((len(self.underlyings),))

        match = "All items in 'contracts' must be unique."
        with pytest.raises(ValueError, match=match):
            MockPortfolioSpace([Cash(), Cash()])

    def test_make_rebalancing_request(self):
        class MockPortfolioSpace(PortfolioSpace, Discrete):
            def __init__(self, contracts, allocations):
                PortfolioSpace.__init__(self, contracts)
                Discrete.__init__(self, n=len(allocations))

            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.ones((len(self.contracts),))

        space = MockPortfolioSpace([Cash(), ETF("IEF")], [[1, 0], [0, 1]])
        action = 0

        actual = space.make_rebalancing_request(action)
        expected = Rebalancing(
            space.contracts, np.ones((2,)), 'weight', True, True, 0, datetime.now()
        )
        assert isinstance(actual.allocation, Weights)
        assert actual.allocation == expected.allocation
        assert actual.absolute is expected.absolute
        assert actual.fractional is expected.fractional
        assert actual.margin == expected.margin

    def test_make_rebalancing_request_if_shortselling_shortable(self):
        class MockPortfolioSpace(PortfolioSpace, Discrete):
            def __init__(self, contracts, allocations):
                self.allocations = allocations
                PortfolioSpace.__init__(self, contracts)
                Discrete.__init__(self, n=len(allocations))

            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.array(self.allocations[action])

        space = MockPortfolioSpace([Cash(), ES(2019, 6)], [[1.1, -0.1]])
        action = 0
        space.make_rebalancing_request(action)  # does not raise

    def test_make_rebalancing_request_raises_if_invalid_action(self):
        class MockPortfolioSpace(PortfolioSpace, Discrete):
            def __init__(self, contracts, allocations):
                PortfolioSpace.__init__(self, contracts)
                Discrete.__init__(self, n=len(allocations))

            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.ones((len(self.contracts),))

        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0, 1]]
        space = MockPortfolioSpace(contracts, allocations)
        action = 42
        with pytest.raises(ValueError):
            # Only 0 and 1 are valid (2 actions)
            space.make_rebalancing_request(action)

    def test_raises_if_margin_is_negative(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.ones((len(self.contracts),))

        with pytest.raises(ValueError):
            MockPortfolioSpace([Cash(), ETF("IEF")], margin=-0.01)

    def test_underlyings(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return np.ones((len(self.contracts),))

            def __contains__(self, item):
                # NotImplementedError otherwise.
                return True


        contracts = [Cash(), ETF("IEF"), FutureChain(ES, "2018-12", "2019-07")]
        space = MockPortfolioSpace(contracts)
        np.testing.assert_equal(
            actual=space.contracts,
            desired=np.array(
                [Cash(), ETF("IEF"), FutureChain(ES, "2018-12", "2019-07")]
            ),
        )

    def test_with_leading_future(self):
        # This is actually an integration test.
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return action

            def __contains__(self, item):
                # NotImplementedError otherwise.
                return True

        class ES(Future):
            multiplier = 50.0
            freq = "Q-DEC"

            def _get_expiry_date(self, year: int, month: int) -> datetime:
                return datetime(year, month, 18)

            def _get_last_trading_date(self, expiry: datetime) -> datetime:
                return datetime(expiry.year, expiry.month, 8)

            def margin_requirement(self) -> float:
                return 0.1

        cash = Cash()
        leading_future = FutureChain(ES, "2019-01-01", "2019-12-31")
        contracts = [cash, leading_future]
        space = MockPortfolioSpace(contracts)
        action = np.array([0.25, 0.75])
        request = space.make_rebalancing_request(action)
        assert request.allocation == {
            leading_future.static_hashing(): 0.75
        }

    def test_cash_is_not_added_to_contracts_if_missing(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return action

        contracts = [ETF("SPY"), ETF("IEF")]
        space = MockPortfolioSpace(contracts)
        assert space.contracts == [ETF("SPY"), ETF("IEF")]

    def test_make_rebalancing_request_when_cash_missing_from_contracts(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return action

            def __contains__(self, item):
                # NotImplementedError otherwise.
                return True

        contracts = [ETF("SPY"), ETF("IEF")]
        space = MockPortfolioSpace(contracts)
        action = np.array([0.25, 0.75])
        rebalancing_request = space.make_rebalancing_request(action)
        assert rebalancing_request.allocation == {
            ETF("SPY"): 0.25,
            ETF("IEF"): 0.75,
        }

    def test_make_rebalancing_request_when_cash_not_first_item_of_contracts(self):
        class MockPortfolioSpace(PortfolioSpace, Space):
            def _make_allocation(self, action: np.ndarray, broker=None):
                return action

            def __contains__(self, item):
                # NotImplementedError otherwise.
                return True

        contracts = [ETF("SPY"), Cash(), ETF("IEF")]
        space = MockPortfolioSpace(contracts)
        action = np.array([0.25, 0.0, 0.75])
        rebalancing_request = space.make_rebalancing_request(action)
        assert rebalancing_request.allocation == {
            ETF("SPY"): 0.25,
            ETF("IEF"): 0.75,
        }


class TestDiscretePortfolio:
    def test_parent_classes(self):
        assert issubclass(DiscretePortfolio, PortfolioSpace)
        assert issubclass(DiscretePortfolio, Discrete)

    def test_initialization(self, mocker):
        mocker.patch.object(PortfolioSpace, "__init__", return_value=None)
        mocker.patch.object(Discrete, "__init__", return_value=None)

        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0.5, 0.5], [0, 1]]
        space = DiscretePortfolio(contracts, allocations)

        PortfolioSpace.__init__.assert_called_once_with(space, contracts, True, True)
        Discrete.__init__.assert_called_once_with(space, n=len(allocations))

    def test_custom_initialization(self):
        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0.5, 0.5], [0, 1]]
        space = DiscretePortfolio(
            contracts, allocations, as_weights=False, fractional=False
        )
        assert space._as_weights is False
        assert space._fractional is False

    def test_allocations_with_different_lengths_raise(self):
        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0.5, 0.5, 0.5], [0, 1]]
        with pytest.raises(ValueError):
            DiscretePortfolio(contracts, allocations)

    def test_contains_when_false(self):
        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0.5, 0.5], [0, 1]]
        space = DiscretePortfolio(contracts, allocations)
        invalid_item = 42
        assert not space.contains(invalid_item)

    def test_contains_when_true(self):
        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0.5, 0.5], [0, 1]]
        space = DiscretePortfolio(contracts, allocations)
        valid_item = 1
        assert space.contains(valid_item)

    def test_sample_returns_an_integer(self):
        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0.5, 0.5], [0, 1]]
        space = DiscretePortfolio(contracts, allocations)
        assert isinstance(space.sample(), int)

    def test_sample_items_from_space(self):
        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0.5, 0.5], [0, 1]]
        space = DiscretePortfolio(contracts, allocations)
        nr_samples = 1000
        for action in [space.sample() for _ in range(nr_samples)]:
            assert action in space

    def test_sample_distribution_is_unbiased(self):
        contracts = [Cash(), ETF("IEF")]
        allocations = [[1, 0], [0.5, 0.5], [0, 1]]
        space = DiscretePortfolio(contracts, allocations)
        nr_samples = 100000
        samples = [space.sample() for _ in range(nr_samples)]
        _, counts = np.unique(samples, return_counts=True, axis=0)
        rel_freq = counts / nr_samples
        assert max(rel_freq) - min(rel_freq) < 0.01

    def test_make_rebalancing_request(self):
        contracts = [Cash(), ETF("IEF"), ETF("GSG")]
        allocations = [[1, 0, 0], [0.7, 0.1, 0.2], [0.3, 0.6, 0.1]]
        space = DiscretePortfolio(contracts, allocations)

        actual = space.make_rebalancing_request(0).allocation
        assert actual == {}

        actual = space.make_rebalancing_request(1).allocation
        assert actual == {ETF('IEF'): 0.1, ETF('GSG'): 0.2}

        actual = space.make_rebalancing_request(2).allocation
        assert actual == {ETF('IEF'): 0.6, ETF('GSG'): 0.1}

    def test_make_rebalancing_request_when_invalid_index(self):
        contracts = [Cash(), ETF("IEF")]
        allocations = [[0, 0], [0.1, 0.2], [0.6, 0.1]]
        space = DiscretePortfolio(contracts, allocations)
        with pytest.raises(ValueError):
            space.make_rebalancing_request(-1)
        with pytest.raises(ValueError):
            space.make_rebalancing_request(3)

    def test_make_rebalancing_request_expands_weights_for_composite_futures(self):
        allocations = [[1, 0, 0], [0.7, 0.1, 0.2], [0.3, 0.6, 0.1]]
        space = DiscretePortfolio(
            contracts=[Cash(), ETF("IEF"), FutureChain(ES, "2018-12", "2019-07")],
            allocations=allocations,
        )
        rebalancing = space.make_rebalancing_request(2)
        assert rebalancing.allocation == {
            ETF("IEF"): 0.6,
            ES(2018, 12): 0.1,
        }

    def test_make_allocation(self):
        contracts = [Cash(), ETF("IEF"), ETF("GSG")]
        allocations = [[1, 0, 0], [0.7, 0.1, 0.2], [0.3, 0.6, 0.1]]
        space = DiscretePortfolio(contracts, allocations)
        assert space._make_allocation(0) == [1, 0, 0]
        assert space._make_allocation(1) == [0.7, 0.1, 0.2]
        assert space._make_allocation(2) == [0.3, 0.6, 0.1]

    def test_make_allocation_with_leading_future(self):
        """It's responsibility of observation_space._make_rebalancing_request to unpack
        the allocation vector."""
        contracts = [Cash(), ETF("IEF"), FutureChain(ES, "2018-12", "2019-07")]
        allocations = [[1, 0, 0], [0.7, 0.1, 0.2], [0.3, 0.6, 0.1]]
        space = DiscretePortfolio(contracts, allocations)
        assert space._make_allocation(0) == [1, 0, 0]
        assert space._make_allocation(1) == [0.7, 0.1, 0.2]
        assert space._make_allocation(2) == [0.3, 0.6, 0.1]


class TestBoxPortfolio:
    def test_parent_classes(self):
        assert issubclass(BoxPortfolio, PortfolioSpace)
        assert issubclass(BoxPortfolio, Box)

    def test_initialization(self, mocker):
        mocker.patch.object(PortfolioSpace, "__init__")
        mocker.patch.object(Box, "__init__")
        mocker.patch.object(Box, "__repr__")

        # To avoid TypeError: __init__() should return None, not 'MagicMock'
        PortfolioSpace.__init__.return_value = None
        Box.__init__.return_value = None
        Box.__repr__.return_value = 'Box'

        contracts = [Cash(), ETF("IEF")]
        space = BoxPortfolio(contracts)

        PortfolioSpace.__init__.assert_called_once_with(
            space, contracts, True, True, 0.0
        )
        Box.__init__.assert_called_once_with(
            space, 0., 1., (len(contracts), ), np.float64
        )

    def test_custom_initialization(self, mocker):
        mocker.patch.object(PortfolioSpace, "__init__")
        mocker.patch.object(Box, "__init__")

        # To avoid TypeError: __init__() should return None, not 'MagicMock'
        PortfolioSpace.__init__.return_value = None
        Box.__init__.return_value = None

        contracts = [Cash(), ETF("IEF")]
        space = BoxPortfolio(
            contracts=contracts,
            low=-1.1,
            high=2.2,
            as_weights=False,
            fractional=False,
            margin=0.02,
        )

        PortfolioSpace.__init__.assert_called_once_with(
            space, contracts, False, False, 0.02
        )
        Box.__init__.assert_called_once_with(space, -1.1, 2.2, (2, ), np.float64)

    def test_contains_is_true_even_when_sum_is_not_one(self):
        contracts = [Cash(), ETF("IEF")]
        space = BoxPortfolio(contracts)
        valid_item = np.array([0.25, 0.5])
        assert space.contains(valid_item)
        assert valid_item in space

    def test_contains_when_true_with_array(self):
        contracts = [Cash(), ETF("IEF")]
        space = BoxPortfolio(contracts)
        valid_item = np.array([0.5, 0.5])
        assert space.contains(valid_item)
        assert valid_item in space

    def test_sample_belongs_to_space(self):
        contracts = [Cash(), ETF("IEF")]
        space = BoxPortfolio(contracts)
        nr_samples = 1000
        for action in [space.sample() for _ in range(nr_samples)]:
            assert action in space

    def test_sample_distribution_is_unbiased(self):
        contracts = [Cash(), ETF("IEF")]
        space = BoxPortfolio(contracts)
        nr_samples = 100000
        samples = [space.sample() for _ in range(nr_samples)]
        mean = np.mean(samples, axis=0)
        np.testing.assert_allclose(mean, np.array([0.5, 0.5]), rtol=0.025)

    def test_make_rebalancing_request(self):
        contracts = [Cash(), ETF("IEF"), ETF("GSG")]
        space = BoxPortfolio(contracts)
        action = np.array([0.7, 0.1, 0.2])
        rebalancing_request = space.make_rebalancing_request(action)
        actual = rebalancing_request.allocation
        expected = {
            ETF("IEF"): 0.1,
            ETF("GSG"): 0.2,
        }
        assert actual == expected

    def test_make_rebalancing_request_action_does_not_have_to_be_array(self):
        contracts = [Cash(), ETF("IEF"), ETF("GSG")]
        space = BoxPortfolio(contracts)
        # Action as list instead of np.ndarray. Does not raise.
        space.make_rebalancing_request([0.25, 0.25, 0.5])

    def test_make_rebalancing_request_expands_weights_for_composite_futures(self):
        contracts = [Cash(), ETF("IEF"), FutureChain(ES, "2018-12", "2019-07")]
        space = BoxPortfolio(contracts)
        action = np.array([0.25, 0.15, 0.6])
        rebalancing = space.make_rebalancing_request(action)
        actual = rebalancing.allocation
        expected = {
            ETF("IEF"): 0.15,
            ES(2018, 12): 0.6,
        }
        assert actual == expected
        assert isinstance(list(actual.keys())[1], ES)



import gym
import gym.utils.seeding
from tradingenv.env import TradingEnv
from tradingenv.spaces import DiscretePortfolio
from tradingenv.contracts import Cash, ETF
from tradingenv.transmitter import AsynchronousTransmitter
import unittest.mock


class TestTradingEnv:
    def get_discrete_env(self, **kwargs):
        env = TradingEnv(
            action_space=DiscretePortfolio(
                allocations=[
                    [0, 0, 1],
                    [0, 0.25, 0.75],
                    [0, 0.5, 0.5],
                    [0, 0.75, 0.25],
                    [0, 1, 0],
                ],
                contracts=[Cash(), ETF("Equities"), ETF("Bonds")],
            ),
            **kwargs,
        )
        return env

    def test_is_subclass_of_gym_env(self):
        assert issubclass(TradingEnv, gym.Env)

    # def test_render_modes_are_specified(self):
    #     assert len(TradingEnv.metadata["render.modes"]) > 0

    @unittest.mock.patch.object(AsynchronousTransmitter, "_create_partitions")
    def test_transmitter_create_partitions_is_called(self, create_partitions):
        self.get_discrete_env()
        create_partitions.assert_called_once()
